#!/usr/bin/env python

import sys
import struct

opcodeSet = {
"WRT": 0,
"NOT": 1,
"WRZ": 2,
"WNZ": 3,
"OR": 4,
"AND": 5,
"XOR": 6,
"BNT": 7,
"GRE": 8,
"GRES": 9,
"EQU": 10,
"ADD": 11,
"SUB": 12,
"BSL": 13,
"BSR": 14,
"INC": 15,
"DEC": 16,
"SLO": 17,
"SRO": 18,
"FNC": 19,
"RET": 20,
"FIL": 21,
"FND": 22,
"CII": 23,
"CIIS": 24,
"MUL": 25,
"MULS": 26,
"DIV": 27,
"DIVS": 28,
"MOD": 29,
"ADDF": 30,
"SUBF": 31,
"MULF": 32,
"DIVF": 33,
"CIF": 34,
"CIFS": 35,
"CFI": 36,
"CFIS": 37,
"CIT": 38,
"CITS": 39,
"CFT": 40,
"CTI": 41,
"CTIS": 42,
"CTF": 43,
"WRTT": 44,
"EQUT": 45,
"FNDT": 46,
"LENT": 47,
"SUBT": 48,
"CONT": 49,
"SDL": 50,
"SDS": 51,
"RAN": 52,
"SIN": 53,
"TWR": 54,
"TRD": 55,
"TSL": 56,
"DCL": 57,
"DWR": 58,
"DRD": 59,
"CKT": 60,
"CTK": 61,
"KIP": 62,
"PKY": 63,
"PTX": 64,
"PSL": 65,
"PFL": 66,
"XSM": 67,
"XDR": 68,
"XDW": 69,
"XAR": 70,
"RNF": 71,
"RGAI": 72,
"RGAN": 73,
"RGAA": 74,
"RCR": 75,
"RRD": 76,
"RWR": 77,
"RGN": 78,
"RGS": 79,
"RDL": 80,
"RRN": 81,
"QUIT": 82,
"CTO": 83,
"CIH": 84,
"CHI": 85,
"ALLO": 86,
"FREE": 87,
"DSC": 88,
"DDX": 89,
"DDY": 90,
"GMLF": 91,
"MNDL": 92,
"POW": 93,
"LOG": 94
}

assemblyOperatorSet = {
"#": 0x00,
"@": 0x10,
"%": 0x20,
"$": 0x30,
"PC": 0x40,
"GS": 0x50,
"LS": 0x60
}

specialTokenNameSet = ["*", "[", "]", ",", ";", "{", "}", "=", "+=", "-=", "*=", "/=", "%=", "&=", "|=", "^=", "<<=", ">>=", "++", "--", "+", "-", "*", "/", "%", "~", "&", "|", "^", "<<", ">>", "!", "&&", "||", "==", "!=", "<", "<=", ">", ">=", "(", ")", ":"]

customTokenCharacterSet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_."

variableTypeBaseNameSet = ["void", "byte", "short", "long", "float", "int", "int2", "data"]

leftUnaryOperatorNameSet = ["-", "!", "~", "*", "&"]

rightUnaryOperatorNameSet = ["++", "--"]

binaryOperatorNameAndPrecedenceSet = {
"*": 3,
"/": 3,
"%": 3,
"+": 4,
"-": 4,
"<<": 5,
">>": 5,
"<": 6,
"<=": 6,
">": 6,
">=": 6,
"==": 7,
"!=": 7,
"&": 8,
"^": 9,
"|": 10,
"&&": 11,
"||": 12,
"=": 13,
"+=": 13,
"-=": 13,
"*=": 13,
"/=": 13,
"%=": 13,
"&=": 13,
"|=": 13,
"^=": 13,
"<<=": 13,
">>=": 13,
",": 14
}

binaryVariableOperatorNameAndPrecedenceSet = {
"*": 3,
"/": 3,
"%": 3,
"+": 4,
"-": 4,
"<<": 5,
">>": 5,
"<": 6,
"<=": 6,
">": 6,
">=": 6,
"==": 7,
"!=": 7,
"&": 8,
"^": 9,
"|": 10,
"&&": 11,
"||": 12
}

functionDestinationVariableName = "__destination__"

pointerSize = 2

class CodeException(Exception):
    def __init__(self, message, lineNumber):
        self.message = message
        self.lineNumber = lineNumber
    def convertToText(self, code):
        output = self.message
        if self.lineNumber == None:
            output += "\n(No associated line number)"
        else:
            output += "\nError on line " + str(self.lineNumber) + ":"
            tempList = code.split("\n")
            index = self.lineNumber - 2
            if index < 0:
                index = 0
            while index < self.lineNumber + 3 and index < len(tempList):
                output += "\n"
                tempText = tempList[index]
                if index == self.lineNumber:
                    output += ">>> "
                output += tempText
                index += 1
        return output

def appendShortToList(list, number):
    list.append(number & 0x00FF)
    list.append((number & 0xFF00) >> 8)

def appendLongToList(list, number):
    list.append(number & 0x000000FF)
    list.append((number & 0x0000FF00) >> 8)
    list.append((number & 0x00FF0000) >> 16)
    list.append((number & 0xFF000000) >> 24)

def appendFloatToList(list, number):
    tempText = struct.pack('!f', number)
    index = len(tempText) - 1
    tempCount = 0
    while tempCount < len(tempText):
        tempCharacter = tempText[index]
        list.append(ord(tempCharacter))
        tempCount += 1
        index -= 1

def appendNumberToList(list, length, number):
    if length == 1:
        list.append(number)
    elif length == 2:
        appendShortToList(list, number)
    elif length == 4:
        if isinstance(number, float):
            appendFloatToList(list, number)
        else:
            appendLongToList(list, number)
    else:
        raise CodeException("Invalid integer length (" + str(length) + ").", None)

def setShortInList(list, index, number):
    list[index] = (number & 0x00FF)
    list[index + 1] = (number & 0xFF00) >> 8

def setLongInList(list, index, number):
    list[index] = (number & 0x000000FF)
    list[index + 1] = (number & 0x0000FF00) >> 8
    list[index + 2] = (number & 0x00FF0000) >> 16
    list[index + 3] = (number & 0xFF000000) >> 24

def setFloatInList(list, index, number):
    tempText = struct.pack('!f', number)
    index2 = len(tempText) - 1
    tempCount = 0
    while tempCount < len(tempText):
        tempCharacter = tempText[index2]
        list[index + tempCount] = (ord(tempCharacter))
        tempCount += 1
        index2 -= 1

def setNumberInList(list, index, length, number):
    if length == 1:
        list[index] = number
    elif length == 2:
        setShortInList(list, index, number)
    elif length == 4:
        if isinstance(number, float):
            setFloatInList(list, index, number)
        else:
            setLongInList(list, index, number)
    else:
        raise CodeException("Invalid integer length (" + str(length) + ").", None)

def getElementByName(list, name):
    index = 0
    while index < len(list):
        tempElement = list[index]
        if tempElement.name == name:
            return tempElement
        index += 1
    return None

def convertHexDigitToNumber(digit):
    output = "0123456789ABCDEF".find(digit.upper())
    if output == -1:
        raise CodeException("Digit is not hexadecimal (" + digit + ").", None)
    return output

nextLabelNumber = 0

def getNextLabelName():
    global nextLabelNumber
    output = "__label" + str(nextLabelNumber) + "__"
    nextLabelNumber += 1
    return output

class Token:
    def __init__(self, name, lineNumber):
        self.name = name
        self.lineNumber = lineNumber
    def __repr__(self):
        return self.name
    def isCustom(self):
        tempCharacter = self.name[0]
        if tempCharacter == "\"" or tempCharacter == "\'":
            return True
        index = 0
        while index < len(self.name):
            tempCharacter = self.name[index]
            if not tempCharacter in customTokenCharacterSet:
                return False
            index += 1
        return True

class VariableTypeBase:
    def __init__(self, isSigned, name):
        self.isSigned = isSigned
        self.name = name
    def __repr__(self):
        if self.isSigned:
            return "signed " + self.name
        else:
            return self.name
    def copy(self):
        return VariableTypeBase(self.isSigned, self.name)
    def getSize(self):
        if self.name == "void":
            return 0
        if self.name == "byte":
            return 1
        if self.name == "short":
            return 2
        if self.name == "long":
            return 4
        if self.name == "float":
            return 4
        if self.name == "int":
            return -1
        if self.name == "data":
            return -2
    def isEqualTo(self, variableTypeBase):
        # For our purposes, we don't care if the sign is different.
        return self.name == variableTypeBase.name

class OperationType:
    def __init__(self, name, precedence):
        self.name = name
        self.precedence = precedence
        self.metadata = None
    def copy(self):
        output = OperationType(self.name, self.precedence)
        output.metadata = self.metadata
        return output
    def __repr__(self):
        return self.name + " " + str(self.precedence) + " " + str(self.metadata)

def leftUnaryCompose(operationType, operand):
    if isinstance(operand, OperationSubstatement):
        return operand.leftUnaryCompose(operationType)
    else:
        tempOperandList = [operand]
        return OperationSubstatement(operationType, tempOperandList)

def leftBinaryCompose(operand1, operationType, operand2):
    if isinstance(operand2, OperationSubstatement):
        return operand2.leftBinaryCompose(operand1, operationType)
    else:
        tempOperandList = [operand1, operand2]
        return OperationSubstatement(operationType, tempOperandList)

class OperationSubstatement:
    def __init__(self, operationType, operandList):
        self.operationType = operationType
        self.operandList = operandList
        self.parentCodeBlockStatement = None
        self.parentStatement = None
    def __repr__(self):
        return str(self.operationType) + " " + str(self.operandList)
    def copy(self):
        tempOperandList = []
        index = 0
        while index < len(self.operandList):
            tempOperand = self.operandList[index]
            if isinstance(tempOperand, OperationSubstatement):
                tempOperandList.append(tempOperand.copy())
            else:
                tempOperandList.append(tempOperand)
            index += 1
        output = OperationSubstatement(self.operationType.copy(), tempOperandList)
        output.parentCodeBlockStatement = self.parentCodeBlockStatement
        output.parentStatement = self.parentStatement
        return output
    def leftUnaryCompose(self, operationType):
        if operationType.precedence < self.operationType.precedence:
            tempOperand = self.operandList[0]
            self.operandList[0] = leftUnaryCompose(operationType, tempOperand)
            return self
        else:
            tempOperandList = [self]
            return OperationSubstatement(operationType, tempOperandList)
    def leftBinaryCompose(self, operand, operationType):
        if not operationType.precedence > self.operationType.precedence:
            tempOperand = self.operandList[0]
            self.operandList[0] = leftBinaryCompose(operand, operationType, tempOperand)
            return self
        else:
            tempOperandList = [operand, self]
            return OperationSubstatement(operationType, tempOperandList)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        self.parentCodeBlockStatement = parentCodeBlockStatement
        self.parentStatement = parentStatement
        index = 0
        while index < len(self.operandList):
            tempOperand = self.operandList[index]
            if isinstance(tempOperand, OperationSubstatement):
                tempOperand.resolveParentRelationships(parentCodeBlockStatement, parentStatement)
            index += 1
    def getIntegerValue(self):
        try:
            if len(self.operandList) == 1:
                if self.operationType.name == "-":
                    tempNumber = getIntegerOperandValue(self.operandList[0])
                    return -tempNumber
                elif self.operationType.name == "!":
                    tempNumber = getIntegerOperandValue(self.operandList[0])
                    return not tempNumber
                elif self.operationType.name == "~":
                    tempNumber = getIntegerOperandValue(self.operandList[0])
                    return ~tempNumber
                elif self.operationType.name == "conversion":
                    tempNumber = getIntegerOperandValue(self.operandList[0])
                    # Too lazy to trim number of bytes.
                    return tempNumber
                else:
                    raise CodeException("Invalid operation for calculating constant integer value (" + self.operationType.name + ").", self.parentStatement.lineNumber)
            else:
                if self.operationType.name == "*":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 * tempNumber2
                elif self.operationType.name == "/":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return int(tempNumber1 / tempNumber2)
                elif self.operationType.name == "%":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 % tempNumber2
                elif self.operationType.name == "+":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 + tempNumber2
                elif self.operationType.name == "-":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 - tempNumber2
                elif self.operationType.name == "<<":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 << tempNumber2
                elif self.operationType.name == ">>":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 >> tempNumber2
                elif self.operationType.name == "<":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 < tempNumber2
                elif self.operationType.name == "<=":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 <= tempNumber2
                elif self.operationType.name == ">":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 > tempNumber2
                elif self.operationType.name == ">=":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 >= tempNumber2
                elif self.operationType.name == "==":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 == tempNumber2
                elif self.operationType.name == "!=":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 != tempNumber2
                elif self.operationType.name == "&":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 & tempNumber2
                elif self.operationType.name == "^":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 ^ tempNumber2
                elif self.operationType.name == "|":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 | tempNumber2
                elif self.operationType.name == "&&":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 and tempNumber2
                elif self.operationType.name == "||":
                    tempNumber1 = getIntegerOperandValue(self.operandList[0])
                    tempNumber2 = getIntegerOperandValue(self.operandList[1])
                    return tempNumber1 or tempNumber2
                else:
                    raise CodeException("Invalid operation for calculating constant integer value (" + self.operationType.name + ").", self.parentStatement.lineNumber)
        except CodeException as exception:
            exception.lineNumber = self.parentStatement.lineNumber
            raise exception
    def stripPointerLevel(self):
        if self.operationType.name == "arrayAccessor" or (self.operationType.name == "*" and len(self.operandList) == 1):
            return (self.operandList[0], True)
        elif self.operationType.name == "identity":
            return stripPointerLevel(self.operandList[0])
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def getPointerLevel(self):
        if self.operationType.name == "arrayAccessor" or (self.operationType.name == "*" and len(self.operandList) == 1):
            return 1 + getPointerLevel(self.operandList[0])
        elif self.operationType.name == "identity":
            return getPointerLevel(self.operandList[0])
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def getVariableDeclarationSize(self, variableTypeBase):
        if self.operationType.name == "arrayAccessor":
            tempTuple = getVariableDeclarationSize(self.operandList[0], variableTypeBase)
            tempLength = 0
            if len(self.operandList) > 1:
                try:
                    tempLength = getIntegerOperandValue(self.operandList[1])
                except CodeException as exception:
                    exception.lineNumber = self.parentStatement.lineNumber
                    raise exception
            tempSize = pointerSize + tempLength * tempTuple[0]
            return (tempSize, True)
        elif self.operationType.name == "*" and len(self.operandList) == 1:
            tempTuple = getVariableDeclarationSize(self.operandList[0], variableTypeBase)
            if tempTuple[1]:
                tempSize = pointerSize + tempTuple[0]
                return (tempSize, True)
            else:
                return (pointerSize, False)
        elif self.operationType.name == "identity":
            return getVariableDeclarationSize(self.operandList[0], variableTypeBase)
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def getVariableDeclarationName(self):
        if self.operationType.name == "arrayAccessor" or (self.operationType.name == "*" and len(self.operandList) == 1) or self.operationType.name == "identity":
            return getVariableDeclarationName(self.operandList[0])
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def setVariableDeclarationName(self, name):
        if self.operationType.name == "arrayAccessor" or (self.operationType.name == "*" and len(self.operandList) == 1) or self.operationType.name == "identity":
            self.operandList[0] = setVariableDeclarationName(self.operandList[0], name)
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def getVariablePassSize(self, variableTypeBase):
        if self.operationType.name == "arrayAccessor" or (self.operationType.name == "*" and len(self.operandList) == 1):
            return pointerSize
        elif self.operationType.name == "identity":
            return getVariablePassSize(self.operandList[0], variableTypeBase)
        else:
            raise CodeException("Invalid operation for variable declaration (" + self.operationType.name + ").", self.parentStatement.lineNumber)
    def convertToBytecode(self, bytecode, constantData):
        try:
            tempDictionary = {"=": "WRT"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertType(tempReturnValueLocation1.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.free()
                return tempReturnValueLocation1
            tempDictionary = {"+=": "ADD", "-=": "SUB"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertType(tempReturnValueLocation1.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsFloat():
                    bytecode.append(opcodeSet[tempOpcodeName + "F"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.free()
                return tempReturnValueLocation1
            tempDictionary = {"*=": "MUL", "/=": "DIV"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertType(tempReturnValueLocation1.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsFloat():
                    bytecode.append(opcodeSet[tempOpcodeName + "F"])
                else:
                    if tempReturnValueLocation1.getIsSigned():
                        bytecode.append(opcodeSet[tempOpcodeName + "S"])
                    else:
                        bytecode.append(opcodeSet[tempOpcodeName])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.free()
                return tempReturnValueLocation1
            tempDictionary = {"%=": "MOD", "&=": "AND", "|=": "OR", "^=": "XOR"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertType(tempReturnValueLocation1.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.free()
                return tempReturnValueLocation1
            tempDictionary = {"<<=": "BSL", ">>=": "BSR"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertToByte(self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.free()
                return tempReturnValueLocation1
            tempDictionary = {"++": "INC", "--": "DEC"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation.getSize())
                tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation
            tempDictionary = {"+": "ADD", "-": "SUB"}
            if self.operationType.name in tempDictionary and len(self.operandList) > 1:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsFloat():
                    bytecode.append(opcodeSet[tempOpcodeName + "F"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                if tempReturnValueLocation2.variableDeclarationSubstatement.variableTypeBase.isSigned:
                    tempReturnValueLocation3.variableDeclarationSubstatement.variableTypeBase.isSigned = True
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"*": "MUL", "/": "DIV"}
            if self.operationType.name in tempDictionary and len(self.operandList) > 1:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsFloat():
                    bytecode.append(opcodeSet[tempOpcodeName + "F"])
                else:
                    if tempReturnValueLocation1.getIsSigned() or tempReturnValueLocation2.getIsSigned():
                        bytecode.append(opcodeSet[tempOpcodeName + "S"])
                    else:
                        bytecode.append(opcodeSet[tempOpcodeName])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                if tempReturnValueLocation2.variableDeclarationSubstatement.variableTypeBase.isSigned:
                    tempReturnValueLocation3.variableDeclarationSubstatement.variableTypeBase.isSigned = True
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"%": "MOD", "&": "AND", "|": "OR", "^": "XOR"}
            if self.operationType.name in tempDictionary and len(self.operandList) > 1:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                if tempReturnValueLocation2.variableDeclarationSubstatement.variableTypeBase.isSigned:
                    tempReturnValueLocation3.variableDeclarationSubstatement.variableTypeBase.isSigned = True
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"-": "SUB"}
            if self.operationType.name in tempDictionary and len(self.operandList) < 2:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                if tempReturnValueLocation1.getIsFloat():
                    bytecode.append(opcodeSet[tempOpcodeName + "F"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                bytecode.append(tempReturnValueLocation1.getSize())
                appendNumberToList(bytecode, tempReturnValueLocation1.getSize(), 0)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                tempReturnValueLocation2.variableDeclarationSubstatement.variableTypeBase.isSigned = True
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation2
            tempDictionary = {"~": "NOT"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation2
            tempDictionary = {"<<": "BSL", ">>": "BSR"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertToByte(self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"!": "BNT"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation2
            tempDictionary = {"&&": "AND", "||": "OR"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                bytecode.append(opcodeSet["GRE"])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation3.getSize())
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                bytecode.append(assemblyOperatorSet["#"] | tempReturnValueLocation1.getSize())
                appendNumberToList(bytecode, tempReturnValueLocation1.getSize(), 0)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"==": "EQU", ">": "GRE"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if (tempReturnValueLocation1.getIsSigned() or tempReturnValueLocation2.getIsSigned()) and tempOpcodeName == "GRE":
                    bytecode.append(opcodeSet[tempOpcodeName + "S"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"!=": "EQU", "<=": "GRE"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if (tempReturnValueLocation1.getIsSigned() or tempReturnValueLocation2.getIsSigned()) and tempOpcodeName == "GRE":
                    bytecode.append(opcodeSet[tempOpcodeName + "S"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                bytecode.append(opcodeSet["BNT"])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation3.getSize())
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"<": "GRE"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsSigned() or tempReturnValueLocation2.getIsSigned():
                    bytecode.append(opcodeSet[tempOpcodeName + "S"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {">=": "GRE"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                assimilateReturnValueLocations(tempReturnValueLocation1, tempReturnValueLocation2, self.parentCodeBlockStatement, bytecode)
                if tempReturnValueLocation1.getIsSigned() or tempReturnValueLocation2.getIsSigned():
                    bytecode.append(opcodeSet[tempOpcodeName + "S"])
                else:
                    bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                bytecode.append(opcodeSet["BNT"])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation3.getSize())
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation3
            tempDictionary = {"&": "ADD"}
            if self.operationType.name in tempDictionary and len(self.operandList) < 2:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, pointerSize)
                tempReturnValueLocation1.appendOffsetToBytecodeAsArgument(bytecode)
                if tempReturnValueLocation1.getIsAbsolute():
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, 0)
                elif tempReturnValueLocation1.getIsGlobal():
                    bytecode.append(assemblyOperatorSet["GS"])
                else:
                    bytecode.append(assemblyOperatorSet["LS"])
                tempVariableDeclarationSubstatement = tempReturnValueLocation1.variableDeclarationSubstatement.shallowCopy()
                tempReturnValueLocation1.free()
                tempVariableDeclarationSubstatement.addPointerLevel()
                tempScratchAllocation = ScratchAllocation(pointerSize, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = ReturnValueLocation(tempVariableDeclarationSubstatement, tempScratchAllocation)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                return tempReturnValueLocation2
            if self.operationType.name == "*" and len(self.operandList) < 2:
                tempReturnValueLocation = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation.dereference()
                return tempReturnValueLocation
            tempDictionary = {"arrayAccessor": "ADD"}
            if self.operationType.name in tempDictionary:
                tempOpcodeName = tempDictionary[self.operationType.name]
                tempReturnValueLocation1 = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2 = convertOperandToBytecode(self.operandList[1], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation2.convertType(tempReturnValueLocation1.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                bytecode.append(opcodeSet[tempOpcodeName])
                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                appendNumberToList(bytecode, pointerSize, tempReturnValueLocation1.getSize())
                tempReturnValueLocation1.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation1.free()
                tempReturnValueLocation2.free()
                tempReturnValueLocation3 = createReturnValueLocationWithScratchAllocation(tempReturnValueLocation1, self.parentCodeBlockStatement)
                tempReturnValueLocation3.appendToBytecodeAsArgument(bytecode)
                tempReturnValueLocation3.dereference()
                return tempReturnValueLocation3
            if self.operationType.name == "conversion":
                tempReturnValueLocation = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                tempReturnValueLocation.convertType(self.operationType.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                return tempReturnValueLocation
            if self.operationType.name == "functionInvocation":
                tempFunctionName = self.operandList[0]
                tempFunctionDeclarationStatement = getDefaultFunctionDeclarationStatementByName(tempFunctionName)
                if tempFunctionDeclarationStatement != None:
                    tempVariableDeclarationSubstatementList = tempFunctionDeclarationStatement.variableDeclarationSubstatementList
                    # The second operand of a function invocation must be a
                    # flattened comma operation substatement.
                    tempOperandList = self.operandList[1].operandList
                    if len(tempOperandList) != len(tempVariableDeclarationSubstatementList):
                        raise CodeException("Wrong number of arguments (" + tempFunctionName + ").", self.parentStatement.lineNumber)
                    returnValueLocationList = []
                    integerSizeList = [-1, -1]
                    integerIsSignedList = [False, False]
                    index = 0
                    while index < len(tempVariableDeclarationSubstatementList):
                        tempReturnValueLocation = convertOperandToBytecode(tempOperandList[index], bytecode, constantData, self.parentCodeBlockStatement)
                        returnValueLocationList.append(tempReturnValueLocation)
                        tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatementList[index]
                        if tempVariableDeclarationSubstatement.getPointerLevel() > 0:
                            tempReturnValueLocation.incrementIndirectInvocationLevel()
                        tempVariableTypeBase = tempVariableDeclarationSubstatement.variableTypeBase
                        if tempVariableTypeBase.name == "int":
                            tempSize = tempReturnValueLocation.getSize()
                            tempSize2 = integerSizeList[tempVariableTypeBase.sizeId]
                            if tempSize > tempSize2:
                                integerSizeList[tempVariableTypeBase.sizeId] = tempSize
                            if tempVariableTypeBase.isSigned:
                                integerIsSignedList[tempVariableTypeBase.sizeId] = True
                        index += 1
                    tempVariableDeclarationSubstatement = tempFunctionDeclarationStatement.variableDeclarationSubstatement
                    tempVariableTypeBase = tempVariableDeclarationSubstatement.variableTypeBase
                    if tempVariableTypeBase.name == "int":
                        if integerSizeList[tempVariableTypeBase.sizeId] < 0:
                            integerSizeList[tempVariableTypeBase.sizeId] = 4
                        if tempVariableTypeBase.isSigned:
                            integerIsSignedList[tempVariableTypeBase.sizeId] = True
                    integerConversionVariableDeclarationSubstatementList = []
                    index = 0
                    while index < len(integerSizeList):
                        tempSize = integerSizeList[index]
                        tempName = None
                        if tempSize == 1:
                            tempName = "byte"
                        elif tempSize == 2:
                            tempName = "short"
                        elif tempSize == 4 or tempSize == -1:
                            tempName = "long"
                        else:
                            raise CodeException("Invalid integer size (" + str(tempSize) + ").", self.parentStatement.lineNumber)
                        tempIsSigned = integerIsSignedList[index]
                        tempVariableTypeBase = VariableTypeBase(tempIsSigned, tempName)
                        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, "__temp__")
                        integerConversionVariableDeclarationSubstatementList.append(tempVariableDeclarationSubstatement)
                        index += 1
                    index = 0
                    while index < len(tempVariableDeclarationSubstatementList):
                        tempReturnValueLocation = returnValueLocationList[index]
                        tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatementList[index]
                        tempVariableTypeBase = tempVariableDeclarationSubstatement.variableTypeBase
                        if tempVariableTypeBase.name == "int":
                            tempVariableDeclarationSubstatement2 = integerConversionVariableDeclarationSubstatementList[tempVariableTypeBase.sizeId]
                            tempReturnValueLocation.convertType(tempVariableDeclarationSubstatement2, self.parentCodeBlockStatement, bytecode)
                        else:
                            tempReturnValueLocation.convertType(tempVariableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                        index += 1
                    tempOpcodeName = defaultFunctionOpcodeNameSet[tempFunctionName]
                    bytecode.append(opcodeSet[tempOpcodeName])
                    hasAddedIntegerSizeArgumentList = [False, False]
                    index = 0
                    while index < len(tempVariableDeclarationSubstatementList):
                        tempReturnValueLocation = returnValueLocationList[index]
                        tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatementList[index]
                        tempVariableTypeBase = tempVariableDeclarationSubstatement.variableTypeBase
                        if tempVariableTypeBase.name == "int":
                            if not hasAddedIntegerSizeArgumentList[tempVariableTypeBase.sizeId]:
                                bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                                appendNumberToList(bytecode, pointerSize, integerSizeList[tempVariableTypeBase.sizeId])
                                hasAddedIntegerSizeArgumentList[tempVariableTypeBase.sizeId] = True
                        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
                        tempReturnValueLocation.free()
                        index += 1
                    tempVariableDeclarationSubstatement = tempFunctionDeclarationStatement.variableDeclarationSubstatement
                    tempVariableTypeBase = tempVariableDeclarationSubstatement.variableTypeBase
                    if tempVariableTypeBase.name == "int":
                        if not hasAddedIntegerSizeArgumentList[tempVariableTypeBase.sizeId]:
                            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                            tempSize = integerSizeList[tempVariableTypeBase.sizeId]
                            appendNumberToList(bytecode, pointerSize, tempSize)
                            hasAddedIntegerSizeArgumentList[tempVariableTypeBase.sizeId] = True
                        tempVariableDeclarationSubstatement = integerConversionVariableDeclarationSubstatementList[tempVariableTypeBase.sizeId]
                    tempSize = tempVariableDeclarationSubstatement.getPassSize()
                    if tempSize > 0:
                        tempScratchAllocation = ScratchAllocation(tempSize, self.parentCodeBlockStatement)
                        tempReturnValueLocation = ReturnValueLocation(tempVariableDeclarationSubstatement, tempScratchAllocation)
                        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
                        return tempReturnValueLocation
                    return None
                tempFunctionDeclarationStatement = self.parentCodeBlockStatement.getFunctionDeclarationStatementByName(tempFunctionName)
                if tempFunctionDeclarationStatement != None:
                    tempVariableDeclarationSubstatementList = tempFunctionDeclarationStatement.variableDeclarationSubstatementList
                    # The second operand of a function invocation must be a
                    # flattened comma operation substatement.
                    tempOperandList = self.operandList[1].operandList
                    if len(tempOperandList) != len(tempVariableDeclarationSubstatementList):
                        raise CodeException("Wrong number of arguments (" + tempFunctionName + ").", self.parentStatement.lineNumber)
                    returnValueLocationList = []
                    index = 0
                    while index < len(tempVariableDeclarationSubstatementList):
                        tempReturnValueLocation = convertOperandToBytecode(tempOperandList[index], bytecode, constantData, self.parentCodeBlockStatement)
                        tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatementList[index]
                        tempReturnValueLocation.convertType(tempVariableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
                        returnValueLocationList.append(tempReturnValueLocation)
                        index += 1
                    tempReturnValueLocation2 = None
                    tempVariableDeclarationSubstatement = tempFunctionDeclarationStatement.variableDeclarationSubstatement
                    tempSize = tempVariableDeclarationSubstatement.getPassSize()
                    if tempSize > 0:
                        tempScratchAllocation = ScratchAllocation(tempSize, self.parentCodeBlockStatement)
                        tempReturnValueLocation2 = ReturnValueLocation(tempVariableDeclarationSubstatement, tempScratchAllocation)
                    bytecode.append(opcodeSet["FNC"])
                    bytecode.append(assemblyOperatorSet["@"] | pointerSize)
                    tempLabelInvocation = LabelInvocation(len(bytecode), False)
                    tempFunctionDeclarationStatement.infoLabelStatement.addLabelInvocation(tempLabelInvocation)
                    appendNumberToList(bytecode, pointerSize, 0)
                    index = 0
                    while index < len(returnValueLocationList):
                        tempReturnValueLocation = returnValueLocationList[index]
                        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
                        tempReturnValueLocation.free()
                        index += 1
                    if tempSize > 0:
                        tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)
                    return tempReturnValueLocation2
                raise CodeException("Unrecognized function name (" + tempFunctionName + ").", self.parentStatement.lineNumber)
            if self.operationType.name == "arrayLiteral":
                tempLabelStatement = LabelStatement(getNextLabelName(), -1)
                tempLabelStatement.offset = len(constantData)
                tempLabelStatement.isRelativeToConstantData = True
                appendNumberToList(constantData, pointerSize, 0)
                tempLabelStatement2 = LabelStatement(getNextLabelName(), -1)
                tempLabelStatement2.offset = len(constantData)
                tempLabelStatement2.isRelativeToConstantData = True
                tempLabelInvocation = LabelInvocation(tempLabelStatement.offset, True)
                tempLabelStatement2.addLabelInvocation(tempLabelInvocation)
                # The operand of an array literal must be a
                # flattened comma operation substatement.
                tempOperandList = self.operandList[0].operandList
                index = 0
                while index < len(tempOperandList):
                    tempOperand = tempOperandList[index]
                    if tempOperand[0] == "\'":
                        tempValue = ord(tempOperand[1])
                        constantData.append(tempValue)
                    else:
                        if tempOperand[:2] == "0x":
                            tempText = tempOperand[2:]
                            tempValue = 0
                            tempShiftAmount = (len(tempText) - 1) * 4
                            index2 = 0
                            while index2 < len(tempText):
                                tempCharacter = tempText[index2]
                                tempNumber = convertHexDigitToNumber(tempCharacter)
                                tempValue += tempNumber << tempShiftAmount
                                tempShiftAmount -= 4
                                index2 += 1
                            constantData.append(tempValue)
                        else:
                            try:
                                tempValue = int(tempOperand)
                            except ValueError:
                                raise CodeException("Operand is not a byte literal (" + tempOperand + ").", self.parentStatement.lineNumber)
                            constantData.append(tempValue)
                    index += 1
                tempVariableTypeBase = VariableTypeBase(False, "data")
                tempOperationSubstatement = OperationSubstatement(OperationType("*", 2), ["__temp__"])
                tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, tempOperationSubstatement)
                tempVariableDeclarationSubstatement.offset = tempLabelStatement
                tempVariableDeclarationSubstatement.isAbsolute = True
                return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
            if self.operationType.name == "identity":
                tempReturnValueLocation = convertOperandToBytecode(self.operandList[0], bytecode, constantData, self.parentCodeBlockStatement)
                return tempReturnValueLocation
            raise CodeException("Unknown operation type (" + self.operationType.name + ").", self.parentStatement.lineNumber)
        except CodeException as exception:
            exception.lineNumber = self.parentStatement.lineNumber
            raise exception
    def flattenCommas(self):
        tempOperandList = []
        index = 0
        while index < len(self.operandList):
            tempOperand = self.operandList[index]
            shouldAppendOperand = True
            if isinstance(tempOperand, OperationSubstatement):
                if tempOperand.operationType.name == ",":
                    tempOperand.flattenCommas()
                    index2 = 0
                    while index2 < len(tempOperand.operandList):
                        tempOperand2 = tempOperand.operandList[index2]
                        tempOperandList.append(tempOperand2)
                        index2 += 1
                    shouldAppendOperand = False
            if shouldAppendOperand:
                tempOperandList.append(tempOperand)
            index += 1
        self.operandList = tempOperandList

def getIntegerOperandValue(operand):
    if isinstance(operand, OperationSubstatement):
        return operand.getIntegerValue()
    else:
        if operand[0] == "\'":
            return ord(operand[1])
        if operand[:2] == "0x":
            tempText = operand[2:]
            tempValue = 0
            tempShiftAmount = (len(tempText) - 1) * 4
            index = 0
            while index < len(tempText):
                tempCharacter = tempText[index]
                tempNumber = convertHexDigitToNumber(tempCharacter)
                tempValue += tempNumber << tempShiftAmount
                tempShiftAmount -= 4
                index += 1
            return tempValue
        try:
            return int(operand)
        except ValueError:
            raise CodeException("Operand is not an integer literal (" + operand + ").", None)

# Returns a tuple containing the stripped element and whether the pointer level
# was successfully stripped.
def stripPointerLevel(operand):
    if isinstance(operand, OperationSubstatement):
        return operand.stripPointerLevel()
    else:
        return (operand, False)

def getPointerLevel(operand):
    if isinstance(operand, OperationSubstatement):
        return operand.getPointerLevel()
    else:
        return 0

# Returns a tuple containing the size and whether an array was allocated.
def getVariableDeclarationSize(operand, variableTypeBase):
    if isinstance(operand, OperationSubstatement):
        return operand.getVariableDeclarationSize(variableTypeBase)
    else:
        tempSize = variableTypeBase.getSize()
        return (tempSize, False)

def getVariableDeclarationName(operand):
    if isinstance(operand, OperationSubstatement):
        return operand.getVariableDeclarationName()
    else:
        return operand

def setVariableDeclarationName(operand, name):
    if isinstance(operand, OperationSubstatement):
        operand.setVariableDeclarationName(name)
        return operand
    else:
        return name

def getVariablePassSize(operand, variableTypeBase):
    if isinstance(operand, OperationSubstatement):
        return operand.getVariablePassSize(variableTypeBase)
    else:
        return variableTypeBase.getSize()

def convertOperandToBytecode(operand, bytecode, constantData, parentCodeBlockStatement):
    if isinstance(operand, OperationSubstatement):
        return operand.convertToBytecode(bytecode, constantData)
    tempVariableDeclarationSubstatement = parentCodeBlockStatement.getVariableDeclarationSubstatementByName(operand)
    if tempVariableDeclarationSubstatement != None:
        tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatement.copy()
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    tempLabelStatement = getLabelStatementByName(operand)
    if tempLabelStatement != None:
        tempVariableTypeBase = VariableTypeBase(False, "data")
        tempOperationSubstatement = OperationSubstatement(OperationType("*", 2), ["__temp__"])
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, tempOperationSubstatement)
        tempVariableDeclarationSubstatement.constantValue = tempLabelStatement
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    if operand[0] == "\"":
        tempLabelStatement = LabelStatement(getNextLabelName(), -1)
        tempLabelStatement.offset = len(constantData)
        tempLabelStatement.isRelativeToConstantData = True
        appendNumberToList(constantData, pointerSize, 0)
        tempLabelStatement2 = LabelStatement(getNextLabelName(), -1)
        tempLabelStatement2.offset = len(constantData)
        tempLabelStatement2.isRelativeToConstantData = True
        tempLabelInvocation = LabelInvocation(tempLabelStatement.offset, True)
        tempLabelStatement2.addLabelInvocation(tempLabelInvocation)
        index = 1
        while index < len(operand) - 1:
            tempCharacter = operand[index]
            constantData.append(ord(tempCharacter))
            index += 1
        constantData.append(0)
        tempVariableTypeBase = VariableTypeBase(False, "data")
        tempOperationSubstatement = OperationSubstatement(OperationType("*", 2), ["__temp__"])
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, tempOperationSubstatement)
        tempVariableDeclarationSubstatement.offset = tempLabelStatement
        tempVariableDeclarationSubstatement.isAbsolute = True
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    if operand[0] == "\'":
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(VariableTypeBase(False, "byte"), "__temp__")
        tempVariableDeclarationSubstatement.constantValue = ord(operand[1])
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    if operand.find(".") != -1:
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(VariableTypeBase(False, "float"), "__temp__")
        tempVariableDeclarationSubstatement.constantValue = float(operand)
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    if operand[:2] == "0x":
        tempText = operand[2:]
        tempTypeBaseName = "byte"
        if len(tempText) > 2 and len(tempText) < 5:
            tempTypeBaseName = "short"
        elif len(tempText) > 4:
            tempTypeBaseName = "long"
        tempValue = 0
        tempShiftAmount = (len(tempText) - 1) * 4
        index = 0
        while index < len(tempText):
            tempCharacter = tempText[index]
            tempNumber = convertHexDigitToNumber(tempCharacter)
            tempValue += tempNumber << tempShiftAmount
            tempShiftAmount -= 4
            index += 1
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(VariableTypeBase(False, tempTypeBaseName), "__temp__")
        tempVariableDeclarationSubstatement.constantValue = tempValue
        return ReturnValueLocation(tempVariableDeclarationSubstatement, None)
    try:
        tempValue = int(operand)
    except ValueError:
        raise CodeException("Operand is not an integer literal (" + operand + ").", None)
    tempTypeBaseName = "byte"
    if tempValue > 255 and tempValue < 65536:
        tempTypeBaseName = "short"
    elif tempValue > 65535:
        tempTypeBaseName = "long"
    tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(VariableTypeBase(False, tempTypeBaseName), "__temp__")
    tempVariableDeclarationSubstatement.constantValue = tempValue
    return ReturnValueLocation(tempVariableDeclarationSubstatement, None)

class VariableDeclarationSubstatement:
    def __init__(self, variableTypeBase, operationSubstatement):
        self.variableTypeBase = variableTypeBase
        self.operationSubstatement = operationSubstatement
        self.constantValue = None
        self.offset = None
        self.isInConstantData = False
        self.isAbsolute = False
        self.indirectInvocationLevel = 0
        self.parentCodeBlockStatement = None
        self.parentStatement = None
    def __repr__(self):
        return str(self.variableTypeBase) + " " + str(self.operationSubstatement)
    def copy(self):
        tempOperationSubstatement = self.operationSubstatement
        if isinstance(tempOperationSubstatement, OperationSubstatement):
            tempOperationSubstatement = tempOperationSubstatement.copy()
        output = VariableDeclarationSubstatement(self.variableTypeBase.copy(), tempOperationSubstatement)
        output.constantValue = self.constantValue
        output.offset = self.offset
        output.isInConstantData = self.isInConstantData
        output.isAbsolute = self.isAbsolute
        output.indirectInvocationLevel = self.indirectInvocationLevel
        output.parentCodeBlockStatement = self.parentCodeBlockStatement
        output.parentStatement = self.parentStatement
        return output
    def shallowCopy(self):
        tempOperationSubstatement = self.operationSubstatement
        if isinstance(tempOperationSubstatement, OperationSubstatement):
            tempOperationSubstatement = tempOperationSubstatement.copy()
        return VariableDeclarationSubstatement(self.variableTypeBase.copy(), tempOperationSubstatement)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        self.parentCodeBlockStatement = parentCodeBlockStatement
        self.parentStatement = parentStatement
        if isinstance(self.operationSubstatement, OperationSubstatement):
            self.operationSubstatement.resolveParentRelationships(parentCodeBlockStatement, parentStatement)
    def getDeclarationSize(self):
        if self.indirectInvocationLevel > 0:
            return (pointerSize, False)
        else:
            return getVariableDeclarationSize(self.operationSubstatement, self.variableTypeBase)
    def getPassSize(self):
        return getVariablePassSize(self.operationSubstatement, self.variableTypeBase)
    def stripPointerLevel(self):
        tempTuple = stripPointerLevel(self.operationSubstatement)
        if not tempTuple[1]:
            raise CodeException("Could not strip pointer level.", self.parentStatement.lineNumber)
        self.operationSubstatement = tempTuple[0]
    def addPointerLevel(self):
        tempOperationType = OperationType("*", 2)
        self.operationSubstatement = OperationSubstatement(tempOperationType, [self.operationSubstatement])
    def getName(self):
        return getVariableDeclarationName(self.operationSubstatement)
    def setName(self, name):
        self.operationSubstatement = setVariableDeclarationName(self.operationSubstatement, name)
    def getIsGlobal(self):
        if self.parentCodeBlockStatement != None:
            return self.parentCodeBlockStatement.getIsGlobal()
        return False
    def getPointerLevel(self):
        return getPointerLevel(self.operationSubstatement)
    def getReferencedSize(self):
        if self.getPointerLevel() < 1:
            return 0
        if self.getPointerLevel() > 1:
            return pointerSize
        return self.variableTypeBase.getSize()
    def isEqualTo(self, variableDeclarationSubstatement):
        if not self.variableTypeBase.isEqualTo(variableDeclarationSubstatement.variableTypeBase):
            return False
        return self.getPointerLevel() == variableDeclarationSubstatement.getPointerLevel()

class Statement:
    def __init__(self, lineNumber):
        self.lineNumber = lineNumber
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        self.parentCodeBlockStatement = parentCodeBlockStatement
        self.parentStatement = parentStatement
    def addAuxiliaryLabelStatements(self, index):
        return index + 1
    def resolveDeclarations(self):
        pass
    def determineVariableOffsets(self):
        pass
    def convertToBytecode(self, bytecode, constantData):
        pass

class StatementWithCodeBlockStatement(Statement):
    def __init__(self, codeBlockStatement, lineNumber):
        Statement.__init__(self, lineNumber)
        self.codeBlockStatement = codeBlockStatement
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        Statement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        self.codeBlockStatement.resolveParentRelationships(parentCodeBlockStatement, self)
    def addAuxiliaryLabelStatements(self, index):
        self.codeBlockStatement.addAuxiliaryLabelStatements(0)
        return index + 1
    def resolveDeclarations(self):
        self.codeBlockStatement.resolveDeclarations()
    def determineVariableOffsets(self):
        self.codeBlockStatement.determineVariableOffsets()
    def convertToBytecode(self, bytecode, constantData):
        self.codeBlockStatement.convertToBytecode(bytecode, constantData)

class StatementWithOperationSubstatement(Statement):
    def __init__(self, operationSubstatement, lineNumber):
        Statement.__init__(self, lineNumber)
        self.operationSubstatement = operationSubstatement
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        Statement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        if isinstance(self.operationSubstatement, OperationSubstatement):
            self.operationSubstatement.resolveParentRelationships(parentCodeBlockStatement, self)
    def convertToBytecode(self, bytecode, constantData):
        global scratchAllocationList
        scratchAllocationList = []
        try:
            return convertOperandToBytecode(self.operationSubstatement, bytecode, constantData, self.parentCodeBlockStatement)
        except CodeException as exception:
            exception.lineNumber = self.lineNumber
            raise exception

class StatementWithVariableDeclarationSubstatement(Statement):
    def __init__(self, variableDeclarationSubstatement, lineNumber):
        Statement.__init__(self, lineNumber)
        self.variableDeclarationSubstatement = variableDeclarationSubstatement
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        Statement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        self.variableDeclarationSubstatement.resolveParentRelationships(parentCodeBlockStatement, self)
    def convertToBytecode(self, bytecode, constantData):
        tempTuple = self.variableDeclarationSubstatement.getDeclarationSize()
        if tempTuple[0] > self.variableDeclarationSubstatement.getPassSize():
            bytecode.append(opcodeSet["ADD"])
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, pointerSize)
            bytecode.append(assemblyOperatorSet["LS"])
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.variableDeclarationSubstatement.offset + 2)
            bytecode.append(assemblyOperatorSet["$"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.variableDeclarationSubstatement.offset)

class VariableDeclarationStatement(StatementWithVariableDeclarationSubstatement):
    def __init__(self, variableDeclarationSubstatement, lineNumber):
        StatementWithVariableDeclarationSubstatement.__init__(self, variableDeclarationSubstatement, lineNumber)
    def __repr__(self):
        return str(self.variableDeclarationSubstatement)
    def resolveDeclarations(self):
        self.parentCodeBlockStatement.addVariableDeclarationSubstatement(self.variableDeclarationSubstatement)
    def convertToBytecode(self, bytecode, constantData):
        StatementWithVariableDeclarationSubstatement.convertToBytecode(self, bytecode, constantData)

class FunctionDeclarationStatement(StatementWithCodeBlockStatement, StatementWithVariableDeclarationSubstatement):
    def __init__(self, variableDeclarationSubstatement, variableDeclarationSubstatementList, codeBlockStatement, lineNumber):
        StatementWithVariableDeclarationSubstatement.__init__(self, variableDeclarationSubstatement, lineNumber)
        StatementWithCodeBlockStatement.__init__(self, codeBlockStatement, lineNumber)
        self.variableDeclarationSubstatement = variableDeclarationSubstatement
        self.variableDeclarationSubstatementList = variableDeclarationSubstatementList
        self.infoLabelStatement = None
    def __repr__(self):
        return str(self.variableDeclarationSubstatement) + " " + str(self.variableDeclarationSubstatementList) + " " + str(self.codeBlockStatement)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        StatementWithCodeBlockStatement.resolveParentRelationships(self, parentCodeBlockStatement, self)
        StatementWithVariableDeclarationSubstatement.resolveParentRelationships(self, parentCodeBlockStatement, self)
    def addAuxiliaryLabelStatements(self, index):
        self.labelStatementBefore = LabelStatement(getNextLabelName(), 0)
        tempLabelStatementAfter = LabelStatement(getNextLabelName(), 0)
        tempGotoStatement = GotoStatement(tempLabelStatementAfter.name, 0)
        tempGotoStatement.resolveParentRelationships(self.parentCodeBlockStatement, self.parentStatement)
        self.parentCodeBlockStatement.statementList.insert(index, tempGotoStatement)
        self.parentCodeBlockStatement.statementList.insert(index + 1, self.labelStatementBefore)
        self.parentCodeBlockStatement.statementList.insert(index + 3, tempLabelStatementAfter)
        StatementWithCodeBlockStatement.addAuxiliaryLabelStatements(self, index)
        return index + 4
    def resolveDeclarations(self):
        index = 0
        while index < len(self.variableDeclarationSubstatementList):
            tempVariableDeclarationSubstatement = self.variableDeclarationSubstatementList[index]
            tempVariableDeclarationSubstatement = tempVariableDeclarationSubstatement.copy();
            tempVariableDeclarationSubstatement.indirectInvocationLevel += 1
            self.codeBlockStatement.addVariableDeclarationSubstatement(tempVariableDeclarationSubstatement)
            index += 1
        if self.variableDeclarationSubstatement.getPassSize() > 0:
            tempVariableDeclarationSubstatement = self.variableDeclarationSubstatement.copy()
            tempVariableDeclarationSubstatement.setName(functionDestinationVariableName)
            tempVariableDeclarationSubstatement.indirectInvocationLevel += 1
            self.codeBlockStatement.addVariableDeclarationSubstatement(tempVariableDeclarationSubstatement)
        StatementWithCodeBlockStatement.resolveDeclarations(self)
        self.parentCodeBlockStatement.addFunctionDeclarationStatement(self)
    def getName(self):
        return self.variableDeclarationSubstatement.getName()
    def addInfoToConstantData(self, constantData):
        self.infoLabelStatement.offset = len(constantData)
        constantData.append(pointerSize)
        tempLabelInvocation = LabelInvocation(len(constantData), True)
        self.labelStatementBefore.addLabelInvocation(tempLabelInvocation)
        appendNumberToList(constantData, pointerSize, 0)
        appendNumberToList(constantData, pointerSize, self.codeBlockStatement.getSize())
        tempAmount = len(self.variableDeclarationSubstatementList)
        tempHasReturnValue = False
        tempHasReturnValue = (self.variableDeclarationSubstatement.getPassSize() > 0)
        if tempHasReturnValue:
            tempAmount += 1
        constantData.append(tempAmount)
    def determineVariableOffsets(self):
        StatementWithCodeBlockStatement.determineVariableOffsets(self)
    def convertToBytecode(self, bytecode, constantData):
        StatementWithCodeBlockStatement.convertToBytecode(self, bytecode, constantData)

class OperationStatement(StatementWithOperationSubstatement):
    def __init__(self, operationSubstatement, lineNumber):
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
    def __repr__(self):
        return str(self.operationSubstatement)
    def convertToBytecode(self, bytecode, constantData):
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        if tempReturnValueLocation != None:
            tempReturnValueLocation.free()

class VariableInitializationStatement(StatementWithVariableDeclarationSubstatement, StatementWithOperationSubstatement):
    def __init__(self, variableDeclarationSubstatement, operationSubstatement, lineNumber):
        StatementWithVariableDeclarationSubstatement.__init__(self, variableDeclarationSubstatement, lineNumber)
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
    def __repr__(self):
        return str(self.variableDeclarationSubstatement) + " = " + str(self.operationSubstatement)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        StatementWithVariableDeclarationSubstatement.resolveParentRelationships(self, parentCodeBlockStatement, self)
        StatementWithOperationSubstatement.resolveParentRelationships(self, parentCodeBlockStatement, self)
    def resolveDeclarations(self):
        self.parentCodeBlockStatement.addVariableDeclarationSubstatement(self.variableDeclarationSubstatement)
    def convertToBytecode(self, bytecode, constantData):
        StatementWithVariableDeclarationSubstatement.convertToBytecode(self, bytecode, constantData)
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        tempReturnValueLocation.convertType(self.variableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
        bytecode.append(opcodeSet["WRT"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, tempReturnValueLocation.getSize())
        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
        tempReturnValueLocation.free()
        tempReturnValueLocation2 = ReturnValueLocation(self.variableDeclarationSubstatement, None)
        tempReturnValueLocation2.appendToBytecodeAsArgument(bytecode)

class GotoStatement(StatementWithOperationSubstatement):
    def __init__(self, operationSubstatement, lineNumber):
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
    def __repr__(self):
        return "goto " + str(self.operationSubstatement)
    def convertToBytecode(self, bytecode, constantData):
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        bytecode.append(opcodeSet["WRT"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, pointerSize)
        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
        bytecode.append(assemblyOperatorSet["PC"])

class IfStatement(StatementWithOperationSubstatement, StatementWithCodeBlockStatement):
    def __init__(self, operationSubstatement, codeBlockStatement, lineNumber):
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
        StatementWithCodeBlockStatement.__init__(self, codeBlockStatement, lineNumber)
    def __repr__(self):
        return "if " + str(self.operationSubstatement) + " " + str(self.codeBlockStatement)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        StatementWithOperationSubstatement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        StatementWithCodeBlockStatement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
    def addAuxiliaryLabelStatements(self, index):
        tempElseStatementIsNext = False
        tempLabelStatement = None
        tempGotoStatement = None
        if index < len(self.parentCodeBlockStatement.statementList) - 1:
            tempStatement = self.parentCodeBlockStatement.statementList[index + 1]
            if isinstance(tempStatement, ElseStatement):
                tempElseStatementIsNext = True
                tempLabelStatement = LabelStatement(getNextLabelName(), 0)
                tempGotoStatement = GotoStatement(tempLabelStatement.name, 0)
                tempGotoStatement.resolveParentRelationships(self.parentCodeBlockStatement, self.parentStatement)
        self.labelStatementAfter = LabelStatement(getNextLabelName(), 0)
        self.parentCodeBlockStatement.statementList.insert(index + 1, self.labelStatementAfter)
        output = index + 2
        if tempElseStatementIsNext:
            self.parentCodeBlockStatement.statementList.insert(index + 1, tempGotoStatement)
            self.parentCodeBlockStatement.statementList.insert(index + 4, tempLabelStatement)
            output += 1
        self.codeBlockStatement.addAuxiliaryLabelStatements(0)
        return output
    def resolveDeclarations(self):
        StatementWithCodeBlockStatement.resolveDeclarations(self)
    def determineVariableOffsets(self):
        StatementWithCodeBlockStatement.determineVariableOffsets(self)
    def convertToBytecode(self, bytecode, constantData):
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        bytecode.append(opcodeSet["WRZ"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, tempReturnValueLocation.getSize())
        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
        tempReturnValueLocation.free()
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, pointerSize)
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        tempLabelInvocation = LabelInvocation(len(bytecode), False)
        self.labelStatementAfter.addLabelInvocation(tempLabelInvocation)
        appendNumberToList(bytecode, pointerSize, 0)
        bytecode.append(assemblyOperatorSet["PC"])
        StatementWithCodeBlockStatement.convertToBytecode(self, bytecode, constantData)

class ElseStatement(StatementWithCodeBlockStatement):
    def __init__(self, codeBlockStatement, lineNumber):
        StatementWithCodeBlockStatement.__init__(self, codeBlockStatement, lineNumber)
    def __repr__(self):
        return "else " + str(self.codeBlockStatement)

class WhileStatement(StatementWithOperationSubstatement, StatementWithCodeBlockStatement):
    def __init__(self, operationSubstatement, codeBlockStatement, lineNumber):
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
        StatementWithCodeBlockStatement.__init__(self, codeBlockStatement, lineNumber)
    def __repr__(self):
        return "while " + str(self.operationSubstatement) + " " + str(self.codeBlockStatement)
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        StatementWithOperationSubstatement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        StatementWithCodeBlockStatement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
    def addAuxiliaryLabelStatements(self, index):
        self.labelStatementBefore = LabelStatement(getNextLabelName(), 0)
        self.labelStatementAfter = LabelStatement(getNextLabelName(), 0)
        tempGotoStatement = GotoStatement(self.labelStatementBefore.name, 0)
        tempGotoStatement.resolveParentRelationships(self.parentCodeBlockStatement, self.parentStatement)
        self.parentCodeBlockStatement.statementList.insert(index, self.labelStatementBefore)
        self.parentCodeBlockStatement.statementList.insert(index + 2, tempGotoStatement)
        self.parentCodeBlockStatement.statementList.insert(index + 3, self.labelStatementAfter)
        self.codeBlockStatement.addAuxiliaryLabelStatements(0)
        return index + 4
    def resolveDeclarations(self):
        StatementWithCodeBlockStatement.resolveDeclarations(self)
    def determineVariableOffsets(self):
        StatementWithCodeBlockStatement.determineVariableOffsets(self)
    def convertToBytecode(self, bytecode, constantData):
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        bytecode.append(opcodeSet["WRZ"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, tempReturnValueLocation.getSize())
        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
        tempReturnValueLocation.free()
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, pointerSize)
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        tempLabelInvocation = LabelInvocation(len(bytecode), False)
        self.labelStatementAfter.addLabelInvocation(tempLabelInvocation)
        appendNumberToList(bytecode, pointerSize, 0)
        bytecode.append(assemblyOperatorSet["PC"])
        StatementWithCodeBlockStatement.convertToBytecode(self, bytecode, constantData)

class BreakStatement(Statement):
    def __init__(self, lineNumber):
        Statement.__init__(self, lineNumber)
    def __repr__(self):
        return "break"
    def convertToBytecode(self, bytecode, constantData):
        tempStatement = self.parentStatement
        while not isinstance(tempStatement, WhileStatement):
            if tempStatement.parentStatement == None:
                raise CodeException("Break statement is not inside of a while statement.", self.lineNumber)
            tempStatement = tempStatement.parentStatement
        bytecode.append(opcodeSet["WRT"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, pointerSize)
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        tempLabelInvocation = LabelInvocation(len(bytecode), False)
        tempStatement.labelStatementAfter.addLabelInvocation(tempLabelInvocation)
        appendNumberToList(bytecode, pointerSize, 0)
        bytecode.append(assemblyOperatorSet["PC"])

class ContinueStatement(Statement):
    def __init__(self, lineNumber):
        Statement.__init__(self, lineNumber)
    def __repr__(self):
        return "continue"
    def convertToBytecode(self, bytecode, constantData):
        tempStatement = self.parentStatement
        while not isinstance(tempStatement, WhileStatement):
            if tempStatement.parentStatement == None:
                raise CodeException("Continue statement is not inside of a while statement.", self.lineNumber)
            tempStatement = tempStatement.parentStatement
        bytecode.append(opcodeSet["WRT"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, pointerSize)
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        tempLabelInvocation = LabelInvocation(len(bytecode), False)
        tempStatement.labelStatementBefore.addLabelInvocation(tempLabelInvocation)
        appendNumberToList(bytecode, pointerSize, 0)
        bytecode.append(assemblyOperatorSet["PC"])

class ReturnWithoutValueStatement(Statement):
    def __init__(self, lineNumber):
        Statement.__init__(self, lineNumber)
    def __repr__(self):
        return "returnWithoutValue"
    def convertToBytecode(self, bytecode, constantData):
        bytecode.append(opcodeSet["RET"])

class ReturnWithValueStatement(StatementWithOperationSubstatement):
    def __init__(self, operationSubstatement, lineNumber):
        StatementWithOperationSubstatement.__init__(self, operationSubstatement, lineNumber)
    def __repr__(self):
        return "returnWithValue " + str(self.operationSubstatement)
    def convertToBytecode(self, bytecode, constantData):
        tempVariableDeclarationSubstatement = self.parentCodeBlockStatement.getVariableDeclarationSubstatementByName(functionDestinationVariableName)
        if tempVariableDeclarationSubstatement == None:
            raise CodeException("Function does not return a value.", self.lineNumber)
        tempReturnValueLocation = StatementWithOperationSubstatement.convertToBytecode(self, bytecode, constantData)
        tempReturnValueLocation.convertType(tempVariableDeclarationSubstatement, self.parentCodeBlockStatement, bytecode)
        bytecode.append(opcodeSet["WRT"])
        bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, tempReturnValueLocation.getSize())
        tempReturnValueLocation.appendToBytecodeAsArgument(bytecode)
        bytecode.append(assemblyOperatorSet["@"] | 0x08 | pointerSize)
        bytecode.append(assemblyOperatorSet["$"] | pointerSize)
        appendNumberToList(bytecode, pointerSize, tempVariableDeclarationSubstatement.offset)
        bytecode.append(opcodeSet["RET"])

class LabelInvocation:
    def __init__(self, offset, isRelativeToConstantData):
        self.offset = offset
        self.isRelativeToConstantData = isRelativeToConstantData

labelStatementList = []

class LabelStatement(Statement):
    def __init__(self, name, lineNumber):
        global labelStatementList
        Statement.__init__(self, lineNumber)
        self.name = name
        self.labelInvocationList = []
        self.offset = None
        self.isRelativeToConstantData = False
        labelStatementList.append(self)
    def __repr__(self):
        return "label " + self.name
    def addLabelInvocation(self, labelInvocation):
        self.labelInvocationList.append(labelInvocation)
    def populateLabelInvocations(self, bytecode, constantData):
        if self.offset == None:
            return
        tempAddress = None
        if self.isRelativeToConstantData:
            tempAddress = len(bytecode) + self.offset
        else:
            tempAddress = self.offset
        index = 0
        while index < len(self.labelInvocationList):
            tempLabelInvocation = self.labelInvocationList[index]
            if tempLabelInvocation.isRelativeToConstantData:
                setNumberInList(constantData, tempLabelInvocation.offset, pointerSize, tempAddress)
            else:
                setNumberInList(bytecode, tempLabelInvocation.offset, pointerSize, tempAddress)
            index += 1
    def convertToBytecode(self, bytecode, constantData):
        self.offset = len(bytecode)

def getLabelStatementByName(name):
    return getElementByName(labelStatementList, name)

class CodeBlockStatement(Statement):
    def __init__(self, statementList, lineNumber):
        Statement.__init__(self, lineNumber)
        self.statementList = statementList
        self.variableDeclarationSubstatementList = []
        self.functionDeclarationStatementList = []
        self.variableSpaceSize = 0
        self.scratchSpaceSize = 0
    def __repr__(self):
        return str(self.statementList)
    def getSize(self):
        tempSize = self.scratchSpaceSize
        index = 0
        while index < len(self.statementList):
            tempStatement = self.statementList[index]
            if isinstance(tempStatement, StatementWithCodeBlockStatement):
                if not isinstance(tempStatement, FunctionDeclarationStatement):
                    tempStatement = tempStatement.codeBlockStatement
            if isinstance(tempStatement, CodeBlockStatement):
                tempSize2 = tempStatement.getSize()
                if tempSize2 > tempSize:
                    tempSize = tempSize2
            index += 1
        return self.variableSpaceSize + tempSize
    def getOffset(self):
        if isinstance(self.parentStatement, FunctionDeclarationStatement) or self.parentStatement == None:
            return 0
        else:
            return self.parentCodeBlockStatement.getOffset() + self.parentCodeBlockStatement.variableSpaceSize
    def resolveParentRelationships(self, parentCodeBlockStatement, parentStatement):
        Statement.resolveParentRelationships(self, parentCodeBlockStatement, parentStatement)
        index = 0
        while index < len(self.statementList):
            tempStatement = self.statementList[index]
            tempStatement.resolveParentRelationships(self, self)
            index += 1
    def addAuxiliaryLabelStatements(self, index):
        tempIndex = 0
        while tempIndex < len(self.statementList):
            tempStatement = self.statementList[tempIndex]
            tempIndex = tempStatement.addAuxiliaryLabelStatements(tempIndex)
    def resolveDeclarations(self):
        if isinstance(self.parentStatement, FunctionDeclarationStatement) or self.parentStatement == None:
            # Reserve space for scope information.
            self.variableSpaceSize = 7
        index = 0
        while index < len(self.statementList):
            tempStatement = self.statementList[index]
            tempStatement.resolveDeclarations()
            index += 1
    def addVariableDeclarationSubstatement(self, variableDeclarationSubstatement):
        self.variableDeclarationSubstatementList.append(variableDeclarationSubstatement)
    def addFunctionDeclarationStatement(self, functionDeclarationStatement):
        self.functionDeclarationStatementList.append(functionDeclarationStatement)
    def determineVariableOffsets(self):
        index = 0
        while index < len(self.variableDeclarationSubstatementList):
            tempVariableDeclarationSubstatement = self.variableDeclarationSubstatementList[index]
            tempVariableDeclarationSubstatement.offset = self.getOffset() + self.variableSpaceSize
            tempTuple = tempVariableDeclarationSubstatement.getDeclarationSize()
            self.variableSpaceSize += tempTuple[0]
            index += 1
        index = 0
        while index < len(self.statementList):
            tempStatement = self.statementList[index]
            tempStatement.determineVariableOffsets()
            index += 1
    def getVariableDeclarationSubstatementByName(self, name):
        index = 0
        while index < len(self.variableDeclarationSubstatementList):
            tempVariableDeclarationSubstatement = self.variableDeclarationSubstatementList[index]
            if tempVariableDeclarationSubstatement.getName() == name:
                return tempVariableDeclarationSubstatement
            index += 1
        if self.parentCodeBlockStatement != None:
            return self.parentCodeBlockStatement.getVariableDeclarationSubstatementByName(name)
        return None
    def getFunctionDeclarationStatementByName(self, name):
        index = 0
        while index < len(self.functionDeclarationStatementList):
            tempFunctionDeclarationStatement = self.functionDeclarationStatementList[index]
            if tempFunctionDeclarationStatement.getName() == name:
                return tempFunctionDeclarationStatement
            index += 1
        if self.parentCodeBlockStatement != None:
            return self.parentCodeBlockStatement.getFunctionDeclarationStatementByName(name)
        return None
    def addFunctionInfoToConstantData(self, constantData):
        index = 0
        while index < len(self.functionDeclarationStatementList):
            tempFunctionDeclarationStatement = self.functionDeclarationStatementList[index]
            tempFunctionDeclarationStatement.addInfoToConstantData(constantData)
            index += 1
    def getIsGlobal(self):
        if isinstance(self.parentStatement, FunctionDeclarationStatement):
            return False
        if self.parentCodeBlockStatement == None:
            return True
        return self.parentCodeBlockStatement.getIsGlobal()
    def convertToBytecode(self, bytecode, constantData):
        index = 0
        while index < len(self.statementList):
            tempStatement = self.statementList[index]
            tempStatement.convertToBytecode(bytecode, constantData)
            index += 1
        if isinstance(self.parentStatement, FunctionDeclarationStatement):
            bytecode.append(opcodeSet["RET"])

scratchAllocationList = []

class ScratchAllocation:
    def __init__(self, size, parentCodeBlockStatement):
        self.size = size
        self.parentCodeBlockStatement = parentCodeBlockStatement
        self.offset = self.parentCodeBlockStatement.getOffset() + self.parentCodeBlockStatement.variableSpaceSize
        self.offsetFromScratchSpace = 0
        hasCollision = True
        while hasCollision:
            hasCollision = False
            index = 0
            while index < len(scratchAllocationList):
                tempScratchAllocation = scratchAllocationList[index]
                if self.offset + self.size > tempScratchAllocation.offset and self.offset < tempScratchAllocation.offset + tempScratchAllocation.size:
                    self.offset = tempScratchAllocation.offset + tempScratchAllocation.size
                    self.offsetFromScratchSpace = tempScratchAllocation.offsetFromScratchSpace + tempScratchAllocation.size
                    hasCollision = True
                index += 1
        scratchAllocationList.append(self)
        if parentCodeBlockStatement.scratchSpaceSize < self.offsetFromScratchSpace + self.size:
            parentCodeBlockStatement.scratchSpaceSize = self.offsetFromScratchSpace + self.size
    def free(self):
        scratchAllocationList.remove(self)

class ReturnValueLocation:
    def __init__(self, variableDeclarationSubstatement, scratchAllocation):
        self.variableDeclarationSubstatement = variableDeclarationSubstatement.copy()
        self.scratchAllocation = scratchAllocation
    def getIsSigned(self):
        return self.variableDeclarationSubstatement.variableTypeBase.isSigned
    def setIsSigned(self, isSigned):
        self.variableDeclarationSubstatement.variableTypeBase.isSigned = isSigned
    def getConstantValue(self):
        return self.variableDeclarationSubstatement.constantValue
    def setConstantValue(self, value):
        self.variableDeclarationSubstatement.constantValue = value
    def getIsInConstantData(self):
        return self.variableDeclarationSubstatement.isInConstantData
    def getIndirectInvocationLevel(self):
        return self.variableDeclarationSubstatement.indirectInvocationLevel
    def getIsAbsolute(self):
        if self.getIndirectInvocationLevel() > 0:
            return True
        return self.variableDeclarationSubstatement.isAbsolute
    def getBaseIsAbsolute(self):
        return self.variableDeclarationSubstatement.isAbsolute
    def getIsGlobal(self):
        if self.scratchAllocation != None:
            return False
        return self.variableDeclarationSubstatement.getIsGlobal()
    def getOffset(self):
        if self.scratchAllocation != None:
            return self.scratchAllocation.offset
        return self.variableDeclarationSubstatement.offset
    def appendOffsetToBytecodeAsArgument(self, bytecode):
        tempLevel = self.getIndirectInvocationLevel()
        if tempLevel == 0:
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
        else:
            while tempLevel > 1:
                bytecode.append(assemblyOperatorSet["@"] | 0x08 | pointerSize)
                tempLevel -= 1
            if tempLevel > 0:
                bytecode.append(self.getArgumentPrefix())
        appendNumberToList(bytecode, pointerSize, self.getOffset())
    def getSize(self):
        return self.variableDeclarationSubstatement.getPassSize()
    def getPointerLevel(self):
        return self.variableDeclarationSubstatement.getPointerLevel()
    # Returns 0 if the value is not a pointer.
    def getReferencedSize(self):
        return self.variableDeclarationSubstatement.getReferencedSize()
    def getIndirectInvocationLevel(self):
        return self.variableDeclarationSubstatement.indirectInvocationLevel
    def incrementIndirectInvocationLevel(self):
        self.variableDeclarationSubstatement.indirectInvocationLevel += 1
    def free(self):
        if self.scratchAllocation != None:
            self.scratchAllocation.free()
    def getArgumentPrefix(self):
        if self.getBaseIsAbsolute():
            tempValue = assemblyOperatorSet["@"]
        elif self.getIsGlobal():
            tempValue = assemblyOperatorSet["%"]
        else:
            tempValue = assemblyOperatorSet["$"]
        return tempValue | pointerSize
    def appendToBytecodeAsArgument(self, bytecode):
        tempValue = self.getConstantValue()
        if isinstance(tempValue, LabelStatement):
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            tempLabelInvocation = LabelInvocation(len(bytecode), False)
            tempValue.addLabelInvocation(tempLabelInvocation)
            appendNumberToList(bytecode, pointerSize, 0)
            return
        if tempValue != None:
            bytecode.append(assemblyOperatorSet["#"] | self.getSize())
            if self.getPointerLevel() > 0:
                appendNumberToList(bytecode, pointerSize, tempValue)
            else:
                tempVariableTypeBaseName = self.variableDeclarationSubstatement.variableTypeBase.name
                if tempVariableTypeBaseName == "byte":
                    bytecode.append(tempValue)
                elif tempVariableTypeBaseName == "short":
                    appendShortToList(bytecode, tempValue)
                elif tempVariableTypeBaseName == "long":
                    appendLongToList(bytecode, tempValue)
                elif tempVariableTypeBaseName == "float":
                    appendFloatToList(bytecode, tempValue)
                else:
                    raise CodeException("Invalid variable type base (" + tempVariableTypeBaseName + ").", None)
            return
        tempCount = 0
        while tempCount < self.getIndirectInvocationLevel():
            bytecode.append(assemblyOperatorSet["@"] | 0x08 | pointerSize)
            tempCount += 1
        tempValue = self.getArgumentPrefix()
        bytecode.append(tempValue)
        tempOffset = self.getOffset()
        if isinstance(tempOffset, LabelStatement):
            tempLabelInvocation = LabelInvocation(len(bytecode), False)
            tempOffset.addLabelInvocation(tempLabelInvocation)
            appendNumberToList(bytecode, pointerSize, 0)
        else:
            appendNumberToList(bytecode, pointerSize, tempOffset)
    def convertType(self, variableDeclarationSubstatement, parentCodeBlockStatement, bytecode):
        if self.variableDeclarationSubstatement.isEqualTo(variableDeclarationSubstatement):
            return
        tempConstantValue = self.getConstantValue()
        tempPointerLevel1 = self.getPointerLevel()
        tempPointerLevel2 = variableDeclarationSubstatement.getPointerLevel()
        tempIsSigned1 = self.getIsSigned()
        tempIsSigned2 = variableDeclarationSubstatement.variableTypeBase.isSigned
        tempName1 = self.variableDeclarationSubstatement.variableTypeBase.name
        tempName2 = variableDeclarationSubstatement.variableTypeBase.name
        tempVariableDeclarationSubstatement1IsFloat = (tempName1 == "float")
        tempVariableDeclarationSubstatement2IsFloat = (tempName2 == "float")
        if tempPointerLevel2 > 0:
            if tempPointerLevel1 == 0:
                tempReferencedSize = variableDeclarationSubstatement.getReferencedSize()
                tempShouldMultiplyNumber = (tempReferencedSize > 1)
                if tempConstantValue != None:
                    tempConstantValue = int(tempConstantValue)
                    if tempShouldMultiplyNumber:
                        tempConstantValue *= tempReferencedSize
                        tempShouldMultiplyNumber = False
                    self.setConstantValue(tempConstantValue)
                    self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase
                    while self.getPointerLevel() < tempPointerLevel2:
                        self.variableDeclarationSubstatement.addPointerLevel()
                elif tempName1 == "float":
                    bytecode.append(opcodeSet["CFIS"])
                    self.appendToBytecodeAsArgument(bytecode)
                    self.free()
                    self.variableDeclarationSubstatement = VariableDeclarationSubstatement(variableDeclarationSubstatement.variableTypeBase.copy(), self.variableDeclarationSubstatement.operationSubstatement)
                    while self.getPointerLevel() < tempPointerLevel2:
                        self.variableDeclarationSubstatement.addPointerLevel()
                    self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, self.getSize())
                    self.appendToBytecodeAsArgument(bytecode)
                elif self.getSize() != pointerSize:
                    if tempIsSigned1:
                        bytecode.append(opcodeSet["CIIS"])
                    else:
                        bytecode.append(opcodeSet["CII"])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, self.getSize())
                    self.appendToBytecodeAsArgument(bytecode)
                    self.free()
                    self.variableDeclarationSubstatement = VariableDeclarationSubstatement(variableDeclarationSubstatement.variableTypeBase.copy(), self.variableDeclarationSubstatement.operationSubstatement)
                    while self.getPointerLevel() < tempPointerLevel2:
                        self.variableDeclarationSubstatement.addPointerLevel()
                    self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, self.getSize())
                    self.appendToBytecodeAsArgument(bytecode)
                else:
                    self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
                    while self.getPointerLevel() < tempPointerLevel2:
                        self.variableDeclarationSubstatement.addPointerLevel()
                tempIsSigned1 = self.getIsSigned()
                if tempShouldMultiplyNumber:
                    if tempIsSigned1:
                        bytecode.append(opcodeSet["MULS"])
                    else:
                        bytecode.append(opcodeSet["MUL"])
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, pointerSize)
                    bytecode.append(assemblyOperatorSet["#"] | pointerSize)
                    appendNumberToList(bytecode, pointerSize, variableDeclarationSubstatement.getReferencedSize())
                    self.appendToBytecodeAsArgument(bytecode)
                    self.free()
                    self.variableDeclarationSubstatement = self.variableDeclarationSubstatement.shallowCopy()
                    self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
                    self.appendToBytecodeAsArgument(bytecode)
            else:
                self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
                while self.getPointerLevel() < tempPointerLevel2:
                    self.variableDeclarationSubstatement.addPointerLevel()
                while self.getPointerLevel() > tempPointerLevel2:
                    self.variableDeclarationSubstatement.stripPointerLevel()
            return
        if tempConstantValue != None:
            if tempName2 == "byte":
                self.setConstantValue(int(tempConstantValue) % 256)
            if tempName2 == "short":
                self.setConstantValue(int(tempConstantValue) % 65536)
            if tempName2 == "long":
                self.setConstantValue(int(tempConstantValue))
            if tempName2 == "int" or tempName2 == "data":
                self.setConstantValue(int(tempConstantValue))
            if tempName2 == "float":
                self.setConstantValue(float(tempConstantValue))
            self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
            return
        if tempName1 == "data" or tempName2 == "data":
            self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
            return
        tempVariableDeclarationSubstatement1IsInt = (tempName1 == "byte" or tempName1 == "short" or tempName1 == "long")
        tempVariableDeclarationSubstatement2IsInt = (tempName2 == "byte" or tempName2 == "short" or tempName2 == "long")
        if tempName1 == "int":
            if tempVariableDeclarationSubstatement2IsInt:
                self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
                return
            else:
                raise CodeException("Cannot convert to integer from non integer type.", None)
        if tempName2 == "int":
            if tempVariableDeclarationSubstatement1IsInt:
                self.variableDeclarationSubstatement.variableTypeBase = variableDeclarationSubstatement.variableTypeBase.copy()
                return
            else:
                raise CodeException("Cannot convert to integer from non integer type.", None)
        if tempVariableDeclarationSubstatement1IsInt and tempVariableDeclarationSubstatement2IsInt:
            if tempIsSigned1:
                bytecode.append(opcodeSet["CIIS"])
            else:
                bytecode.append(opcodeSet["CII"])
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.getSize())
            self.appendToBytecodeAsArgument(bytecode)
            self.free()
            self.variableDeclarationSubstatement = VariableDeclarationSubstatement(variableDeclarationSubstatement.variableTypeBase.copy(), self.variableDeclarationSubstatement.operationSubstatement)
            self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.getSize())
            self.appendToBytecodeAsArgument(bytecode)
            return
        if tempVariableDeclarationSubstatement1IsFloat and tempVariableDeclarationSubstatement2IsInt:
            if tempIsSigned2:
                bytecode.append(opcodeSet["CFIS"])
            else:
                bytecode.append(opcodeSet["CFI"])
            self.appendToBytecodeAsArgument(bytecode)
            self.free()
            self.variableDeclarationSubstatement = VariableDeclarationSubstatement(variableDeclarationSubstatement.variableTypeBase.copy(), self.variableDeclarationSubstatement.operationSubstatement)
            self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.getSize())
            self.appendToBytecodeAsArgument(bytecode)
            return
        if tempVariableDeclarationSubstatement1IsInt and tempVariableDeclarationSubstatement2IsFloat:
            if tempIsSigned1:
                bytecode.append(opcodeSet["CIFS"])
            else:
                bytecode.append(opcodeSet["CIF"])
            bytecode.append(assemblyOperatorSet["#"] | pointerSize)
            appendNumberToList(bytecode, pointerSize, self.getSize())
            self.appendToBytecodeAsArgument(bytecode)
            self.free()
            self.variableDeclarationSubstatement = VariableDeclarationSubstatement(variableDeclarationSubstatement.variableTypeBase.copy(), self.variableDeclarationSubstatement.operationSubstatement)
            self.scratchAllocation = ScratchAllocation(self.getSize(), parentCodeBlockStatement)
            self.appendToBytecodeAsArgument(bytecode)
            return
    def convertToByte(self, parentCodeBlockStatement, bytecode):
        tempVariableTypeBase = VariableTypeBase(False, "byte")
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, "__temp__")
        self.convertType(tempVariableDeclarationSubstatement, parentCodeBlockStatement, bytecode)
    def getIsFloat(self):
        return self.variableDeclarationSubstatement.variableTypeBase.name == "float" and self.getPointerLevel() < 1
    def getIsSigned(self):
        return self.variableDeclarationSubstatement.variableTypeBase.isSigned
    def dereference(self):
        self.variableDeclarationSubstatement = self.variableDeclarationSubstatement.copy()
        self.variableDeclarationSubstatement.stripPointerLevel()
        self.variableDeclarationSubstatement.indirectInvocationLevel += 1

def createReturnValueLocationWithScratchAllocation(returnValueLocationOrSize, parentCodeBlockStatement):
    tempScratchAllocation = None
    tempVariableDeclarationSubstatement = None
    if isinstance(returnValueLocationOrSize, ReturnValueLocation):
        tempReturnValueLocation = returnValueLocationOrSize
        tempScratchAllocation = ScratchAllocation(tempReturnValueLocation.getSize(), parentCodeBlockStatement)
        tempVariableDeclarationSubstatement = tempReturnValueLocation.variableDeclarationSubstatement.shallowCopy()
    else:
        tempSize = returnValueLocationOrSize
        tempScratchAllocation = ScratchAllocation(tempSize, parentCodeBlockStatement)
        tempVariableTypeBase = None
        if tempSize == 1:
            tempVariableTypeBase = VariableTypeBase(False, "byte")
        elif tempSize == 2:
            tempVariableTypeBase = VariableTypeBase(False, "short")
        elif tempSize == 4:
            tempVariableTypeBase = VariableTypeBase(False, "long")
        else:
            raise CodeException("Invalid integer size (" + str(tempSize) + ").", None)
        tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, "__temp__")
    return ReturnValueLocation(tempVariableDeclarationSubstatement, tempScratchAllocation) 

def assimilateReturnValueLocations(returnValueLocation1, returnValueLocation2, parentCodeBlockStatement, bytecode):
    if returnValueLocation1.getPointerLevel() > 0 or returnValueLocation1.getSize() > returnValueLocation2.getSize() or returnValueLocation1.getIsFloat():
        returnValueLocation2.convertType(returnValueLocation1.variableDeclarationSubstatement, parentCodeBlockStatement, bytecode)
    else:
        returnValueLocation1.convertType(returnValueLocation2.variableDeclarationSubstatement, parentCodeBlockStatement, bytecode)
    tempIsSigned1 = returnValueLocation1.getIsSigned()
    tempIsSigned2 = returnValueLocation2.getIsSigned()
    if tempIsSigned1 or tempIsSigned2:
        returnValueLocation1.setIsSigned(True)
        returnValueLocation2.setIsSigned(True)

# Returns a tuple containing the character and the index of the next character.
def getEscapedCharacter(text, index):
    tempCharacter = text[index]
    index += 1
    outputCharacter = tempCharacter
    if tempCharacter == "n":
        outputCharacter = "\n"
    elif tempCharacter == "\n":
        outputCharacter = ""
    elif tempCharacter == "t":
        outputCharacter = "\t"
    elif tempCharacter == "x":
        tempCharacter1 = text[index]
        index += 1
        tempCharacter2 = text[index]
        index += 1
        outputCharacter = chr((convertHexDigitToNumber(tempCharacter1) << 4) + convertHexDigitToNumber(tempCharacter2));
    return (outputCharacter, index)

def convertCodeToTokenList(code):
    output = []
    tempLineNumber = 0
    index = 0
    while index < len(code):
        while index < len(code):
            tempCharacter = code[index]
            if tempCharacter == "\n":
                tempLineNumber += 1
            if tempCharacter != " " and tempCharacter != "\t" and tempCharacter != "\n":
                break
            index += 1
        if index > len(code) - 1:
            break
        tempText1 = code[index:(index + 3)]
        tempText2 = code[index:(index + 2)]
        tempCharacter = code[index]
        if tempText1 in specialTokenNameSet:
            output.append(Token(tempText1, tempLineNumber))
            index += len(tempText1)
        elif tempText2 in specialTokenNameSet:
            output.append(Token(tempText2, tempLineNumber))
            index += len(tempText2)
        elif tempCharacter in specialTokenNameSet:
            output.append(Token(tempCharacter, tempLineNumber))
            index += len(tempCharacter)
        elif tempCharacter == "\'":
            tempName = tempCharacter
            index += 1
            tempCharacter = code[index]
            index += 1
            if tempCharacter == "\\":
                tempTuple = getEscapedCharacter(code, index)
                tempName += tempTuple[0]
                index = tempTuple[1]
            else:
                tempName += tempCharacter
            tempCharacter = code[index]
            index += 1
            if tempCharacter != "\'":
                raise CodeException("Missing close apostrophe.", tempLineNumber)
            tempName += tempCharacter
            output.append(Token(tempName, tempLineNumber))
        elif tempCharacter == "\"":
            tempName = tempCharacter
            index += 1
            characterIsEscaped = False
            while True:
                tempCharacter = code[index]
                if tempCharacter == "\n":
                    tempLineNumber += 1
                if characterIsEscaped:
                    tempTuple = getEscapedCharacter(code, index)
                    tempName += tempTuple[0]
                    index = tempTuple[1]
                    characterIsEscaped = False
                else:
                    index += 1
                    if tempCharacter == "\\":
                        characterIsEscaped = True
                    else:
                        tempName += tempCharacter
                        if tempCharacter == "\"":
                            break
                if not index < len(code):
                    raise CodeException("Missing close quotation marks.", None)
            output.append(Token(tempName, tempLineNumber))
        elif customTokenCharacterSet.find(tempCharacter) != -1:
            tempName = tempCharacter
            index += 1
            while index < len(code):
                tempCharacter = code[index]
                if customTokenCharacterSet.find(tempCharacter) == -1:
                    break
                index += 1
                tempName += tempCharacter
            output.append(Token(tempName, tempLineNumber))
        else:
            raise CodeException("Unrecognized symbol (" + tempCharacter + ").", tempLineNumber)
    return output

def convertOperandToFlattenedCommaOperationSubstatement(operand):
    if isinstance(operand, OperationSubstatement):
        if operand.operationType.name == ",":
            operand.flattenCommas()
            return operand
    tempOperationType = OperationType(",", 14)
    tempOperandList = [operand]
    return OperationSubstatement(tempOperationType, tempOperandList)

def getNextOperationSubstatementAfterOperand(tokenList, index, operand, isVariableOperation):
    if not index < len(tokenList):
        return (operand, index)
    tempBinaryOperatorNameAndPrecedenceSet = binaryOperatorNameAndPrecedenceSet
    if isVariableOperation:
        tempBinaryOperatorNameAndPrecedenceSet = binaryVariableOperatorNameAndPrecedenceSet
    tempToken = tokenList[index]
    if tempToken.name == "(" and not isVariableOperation:
        index += 1
        if not index < len(tokenList):
            raise CodeException("Unexpected end of code.", None)
        tempOperationSubstatement = None
        tempToken2 = tokenList[index]
        if tempToken2.name == ")":
            index += 1
            tempOperationType = OperationType("functionInvocation", 1)
            tempOperationType2 = OperationType(",", 14)
            tempOperand = OperationSubstatement(tempOperationType2, [])
            tempOperandList = [operand, tempOperand]
            tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        else:
            tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
            index = tempTuple[1]
            if not index < len(tokenList):
                raise CodeException("Unexpected end of code.", None)
            tempToken2 = tokenList[index]
            index += 1
            if tempToken2.name != ")":
                raise CodeException("Missing close parenthesis.", tempToken.lineNumber)
            tempOperationType = OperationType("functionInvocation", 1)
            tempOperand = tempTuple[0]
            tempOperand = convertOperandToFlattenedCommaOperationSubstatement(tempOperand)
            tempOperandList = [operand, tempOperand]
            tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        return getNextOperationSubstatementAfterOperand(tokenList, index, tempOperationSubstatement, isVariableOperation)
    elif tempToken.name in tempBinaryOperatorNameAndPrecedenceSet:
        index += 1
        tempPrecedence = tempBinaryOperatorNameAndPrecedenceSet[tempToken.name]
        tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
        tempOperationType = OperationType(tempToken.name, tempPrecedence)
        tempOperationSubstatement = leftBinaryCompose(operand, tempOperationType, tempTuple[0])
        return (tempOperationSubstatement, tempTuple[1])
    elif tempToken.name in rightUnaryOperatorNameSet and not isVariableOperation:
        # The only right unary operators are currently ++ and --. We do not
        # want these as variable operations.
        index += 1
        tempOperationType = OperationType(tempToken.name, 1)
        tempOperandList = [operand]
        tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        return getNextOperationSubstatementAfterOperand(tokenList, index, tempOperationSubstatement, isVariableOperation)
    elif tempToken.name == "[":
        index += 1
        tempOperationSubstatement = None
        tempToken2 = tokenList[index]
        if tempToken2.name == "]":
            index += 1
            tempOperationType = OperationType("arrayAccessor", 1)
            tempOperandList = [operand]
            tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        else:
            tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
            index = tempTuple[1]
            if not index < len(tokenList):
                raise CodeException("Unexpected end of code.", None)
            tempToken2 = tokenList[index]
            index += 1
            if tempToken2.name != "]":
                raise CodeException("Missing close bracket.", tempToken.lineNumber)
            tempOperationType = OperationType("arrayAccessor", 1)
            tempOperandList = [operand, tempTuple[0]]
            tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        return getNextOperationSubstatementAfterOperand(tokenList, index, tempOperationSubstatement, isVariableOperation)
    else:
        return (operand, index)

def getNextOperationSubstatement(tokenList, index, isVariableOperation):
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.isCustom():
        return getNextOperationSubstatementAfterOperand(tokenList, index, tempToken.name, isVariableOperation)
    elif tempToken.name in leftUnaryOperatorNameSet:
        tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
        tempOperationType = OperationType(tempToken.name, 2)
        tempOperationSubstatement = leftUnaryCompose(tempOperationType, tempTuple[0])
        return (tempOperationSubstatement, tempTuple[1])
    elif tempToken.name == "(":
        tempToken2 = tokenList[index]
        isSigned = (tempToken2.name == "signed")
        if isSigned or tempToken2.name in variableTypeBaseNameSet:
            index += 1
            if isSigned:
                tempToken2 = tokenList[index]
                index += 1
            tempVariableTypeBase = VariableTypeBase(isSigned, tempToken2.name)
            tempPointerLevel = 0
            while index < len(tokenList):
                tempToken3 = tokenList[index]
                index += 1
                if tempToken3.name == ")":
                    break
                elif tempToken3.name == "*":
                    tempPointerLevel += 1
                else:
                    raise CodeException("Invalid token for conversion type (" + tempToken3.name + ").", tempToken.lineNumber)
            tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
            tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, "__temp__")
            tempCount = 0
            while tempCount < tempPointerLevel:
                tempVariableDeclarationSubstatement.addPointerLevel()
                tempCount += 1
            tempOperationType = OperationType("conversion", 2)
            tempOperationType.variableDeclarationSubstatement = tempVariableDeclarationSubstatement
            tempOperationSubstatement = leftUnaryCompose(tempOperationType, tempTuple[0])
            return (tempOperationSubstatement, tempTuple[1])
        else:
            tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
            index = tempTuple[1]
            if not index < len(tokenList):
                raise CodeException("Unexpected end of code.", None)
            tempToken3 = tokenList[index]
            index += 1
            if tempToken3.name != ")":
                raise CodeException("Missing close parenthesis.", tempToken.lineNumber)
            tempOperationType = OperationType("identity", 0)
            tempOperandList = [tempTuple[0]]
            tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
            return getNextOperationSubstatementAfterOperand(tokenList, index, tempOperationSubstatement, isVariableOperation)
    elif tempToken.name == "{":
        tempTuple = getNextOperationSubstatement(tokenList, index, isVariableOperation)
        index = tempTuple[1]
        if not index < len(tokenList):
            raise CodeException("Unexpected end of code.", None)
        tempToken3 = tokenList[index]
        index += 1
        if tempToken3.name != "}":
            raise CodeException("Missing close curly brace.", tempToken.lineNumber)
        tempOperationType = OperationType("arrayLiteral", 0)
        tempOperand = tempTuple[0]
        tempOperand = convertOperandToFlattenedCommaOperationSubstatement(tempOperand)
        tempOperandList = [tempOperand]
        tempOperationSubstatement = OperationSubstatement(tempOperationType, tempOperandList)
        return getNextOperationSubstatementAfterOperand(tokenList, index, tempOperationSubstatement, isVariableOperation)
    else:
        raise CodeException("Unexpected token (" + tempToken.name + ").", tempToken.lineNumber)

def getNextVariableDeclarationSubstatement(tokenList, index):
    tempToken = tokenList[index]
    index += 1
    tempIsSigned = (tempToken.name == "signed")
    if tempIsSigned:
        tempToken = tokenList[index]
        index += 1
    tempVariableTypeBaseName = None
    tempVariableTypeBaseSizeId = None
    if tempToken.name == "int":
        tempVariableTypeBaseName = "int"
        tempVariableTypeBaseSizeId = 0
    elif tempToken.name == "int2":
        tempVariableTypeBaseName = "int"
        tempVariableTypeBaseSizeId = 1
    else:
        tempVariableTypeBaseName = tempToken.name
    tempVariableTypeBase = VariableTypeBase(tempIsSigned, tempVariableTypeBaseName)
    if tempVariableTypeBaseSizeId != None:
        tempVariableTypeBase.sizeId = tempVariableTypeBaseSizeId
    tempTuple = getNextOperationSubstatement(tokenList, index, True)
    tempVariableDeclarationSubstatement = VariableDeclarationSubstatement(tempVariableTypeBase, tempTuple[0])
    return (tempVariableDeclarationSubstatement, tempTuple[1])

def getNextFunctionDeclarationStatement(tokenList, index, variableDeclarationSubstatement):
    index += 1
    tempVariableDeclarationSubstatementList = []
    tempToken = tokenList[index]
    if tempToken.name == ")":
        index += 1
    else:
        while True:
            tempTuple = getNextVariableDeclarationSubstatement(tokenList, index)
            index = tempTuple[1]
            tempVariableDeclarationSubstatementList.append(tempTuple[0])
            tempToken = tokenList[index]
            index += 1
            if tempToken.name == ")":
                break
            elif tempToken.name != ",":
                raise CodeException("Expected a comma.", tempToken.lineNumber)
    tempTuple = getNextStatement(tokenList, index)
    tempStatement = tempTuple[0]
    tempCodeBlockStatement = tempStatement
    if not isinstance(tempStatement, CodeBlockStatement):
        tempStatementList = [tempStatement]
        tempCodeBlockStatement = CodeBlockStatement(tempStatementList, tempToken.lineNumber)
    tempFunctionDeclarationStatement = FunctionDeclarationStatement(variableDeclarationSubstatement, tempVariableDeclarationSubstatementList, tempCodeBlockStatement, tempToken.lineNumber)
    tempFunctionDeclarationStatement.infoLabelStatement = LabelStatement(getNextLabelName(), -1)
    tempFunctionDeclarationStatement.infoLabelStatement.isRelativeToConstantData = True
    return (tempFunctionDeclarationStatement, tempTuple[1])

def getNextVariableDeclarationStatement(tokenList, index, variableDeclarationSubstatement):
    tempToken = tokenList[index]
    index += 1
    tempVariableDeclarationStatement = VariableDeclarationStatement(variableDeclarationSubstatement, tempToken.lineNumber)
    return (tempVariableDeclarationStatement, index)

def getNextVariableInitializationStatement(tokenList, index, variableDeclarationSubstatement):
    tempStartToken = tokenList[index]
    index += 1
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempVariableInitializationStatement = VariableInitializationStatement(variableDeclarationSubstatement, tempTuple[0], tempToken.lineNumber)
    return (tempVariableInitializationStatement, index)

def getNextGotoStatement(tokenList, index):
    tempStartToken = tokenList[index]
    index += 1
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempStatement = GotoStatement(tempTuple[0], tempToken.lineNumber)
    return (tempStatement, index)

def getNextIfStatement(tokenList, index):
    tempToken = tokenList[index]
    index += 1
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    tempTuple2 = getNextStatement(tokenList, index)
    tempStatement = tempTuple2[0]
    tempCodeBlockStatement = tempStatement
    if not isinstance(tempStatement, CodeBlockStatement):
        tempStatementList = [tempStatement]
        tempCodeBlockStatement = CodeBlockStatement(tempStatementList, tempToken.lineNumber)
    tempStatement = IfStatement(tempTuple[0], tempCodeBlockStatement, tempToken.lineNumber)
    return (tempStatement, tempTuple2[1])

def getNextElseStatement(tokenList, index):
    tempToken = tokenList[index]
    index += 1
    tempTuple = getNextStatement(tokenList, index)
    tempStatement = tempTuple[0]
    tempCodeBlockStatement = tempStatement
    if not isinstance(tempStatement, CodeBlockStatement):
        tempStatementList = [tempStatement]
        tempCodeBlockStatement = CodeBlockStatement(tempStatementList, tempToken.lineNumber)
    tempStatement = ElseStatement(tempCodeBlockStatement, tempToken.lineNumber)
    return (tempStatement, tempTuple[1])

def getNextWhileStatement(tokenList, index):
    tempToken = tokenList[index]
    index += 1
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    tempTuple2 = getNextStatement(tokenList, index)
    tempStatement = tempTuple2[0]
    tempCodeBlockStatement = tempStatement
    if not isinstance(tempStatement, CodeBlockStatement):
        tempStatementList = [tempStatement]
        tempCodeBlockStatement = CodeBlockStatement(tempStatementList, tempToken.lineNumber)
    tempStatement = WhileStatement(tempTuple[0], tempCodeBlockStatement, tempToken.lineNumber)
    return (tempStatement, tempTuple2[1])

def getNextBreakStatement(tokenList, index):
    tempStartToken = tokenList[index]
    index += 1
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempStatement = BreakStatement(tempToken.lineNumber)
    return (tempStatement, index)

def getNextContinueStatement(tokenList, index):
    tempStartToken = tokenList[index]
    index += 1
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempStatement = ContinueStatement(tempToken.lineNumber)
    return (tempStatement, index)

def getNextReturnWithoutValueStatement(tokenList, index):
    tempStartToken = tokenList[index]
    index += 1
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempStatement = ReturnWithoutValueStatement(tempToken.lineNumber)
    return (tempStatement, index)

def getNextReturnWithValueStatement(tokenList, index):
    tempStartToken = tokenList[index]
    index += 1
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempStatement = ReturnWithValueStatement(tempTuple[0], tempToken.lineNumber)
    return (tempStatement, index)

def getNextLabelStatement(tokenList, index):
    tempToken = tokenList[index]
    index += 2
    tempStatement = LabelStatement(tempToken.name, tempToken.lineNumber)
    return (tempStatement, index)

def getNextCodeBlockStatement(tokenList, index):
    index += 1
    tempStatementList = []
    while True:
        if not index < len(tokenList):
            raise CodeException("Missing a close curly brace.", None)
        tempToken = tokenList[index]
        if tempToken.name == "}":
            index += 1
            break
        tempTuple = getNextStatement(tokenList, index)
        index = tempTuple[1]
        tempStatementList.append(tempTuple[0])
    tempStatement = CodeBlockStatement(tempStatementList, tempToken.lineNumber)
    return (tempStatement, index)

def getNextOperationStatement(tokenList, index):
    tempStartToken = tokenList[index]
    tempTuple = getNextOperationSubstatement(tokenList, index, False)
    index = tempTuple[1]
    if not index < len(tokenList):
        raise CodeException("Unexpected end of code.", None)
    tempToken = tokenList[index]
    index += 1
    if tempToken.name != ";":
        raise CodeException("Missing a semicolon.", tempStartToken.lineNumber)
    tempOperationStatement = OperationStatement(tempTuple[0], tempToken.lineNumber)
    return (tempOperationStatement, index)

# Returns a tuple containing a statement and the next index.
def getNextStatement(tokenList, index):
    tempToken = tokenList[index]
    if tempToken.name == "signed" or tempToken.name in variableTypeBaseNameSet:
        tempTuple = getNextVariableDeclarationSubstatement(tokenList, index)
        tempVariableDeclarationSubstatement = tempTuple[0]
        index = tempTuple[1]
        if not index < len(tokenList):
            raise CodeException("Unexpected end of code.", None)
        tempToken2 = tokenList[index]
        if tempToken2.name == "(":
            return getNextFunctionDeclarationStatement(tokenList, index, tempVariableDeclarationSubstatement)
        elif tempToken2.name == "=":
            return getNextVariableInitializationStatement(tokenList, index, tempVariableDeclarationSubstatement)
        elif tempToken2.name == ";":
            return getNextVariableDeclarationStatement(tokenList, index, tempVariableDeclarationSubstatement)
        else:
            raise CodeException("Unexpected token (" + tempToken2.name + ").", tempToken.lineNumber)
    elif tempToken.name == "goto":
        return getNextGotoStatement(tokenList, index)
    elif tempToken.name == "if":
        return getNextIfStatement(tokenList, index)
    elif tempToken.name == "else":
        return getNextElseStatement(tokenList, index)
    elif tempToken.name == "while":
        return getNextWhileStatement(tokenList, index)
    elif tempToken.name == "break":
        return getNextBreakStatement(tokenList, index)
    elif tempToken.name == "continue":
        return getNextContinueStatement(tokenList, index)
    elif tempToken.name == "return":
        tempToken2 = tokenList[index + 1]
        if tempToken2.name == ";":
            return getNextReturnWithoutValueStatement(tokenList, index)
        else:
            return getNextReturnWithValueStatement(tokenList, index)
    elif tempToken.name == "{":
        return getNextCodeBlockStatement(tokenList, index)
    elif index < len(tokenList) - 1 and (tokenList[index + 1]).name == ":":
        return getNextLabelStatement(tokenList, index)
    else:
        return getNextOperationStatement(tokenList, index)

def convertTokenListToStatement(tokenList):
    tempStatementList = []
    index = 0
    while index < len(tokenList):
        tempTuple = getNextStatement(tokenList, index)
        tempStatement = tempTuple[0]
        tempStatementList.append(tempStatement)
        index = tempTuple[1]
    return CodeBlockStatement(tempStatementList, 0)

defaultFunctionDeclarationStatementSet = []

defaultFunctionOpcodeNameSet = {
"write": "WRT",
"fill": "FIL",
"find": "FND",
"convertIntToString": "CIT",
"convertSignedIntToString": "CITS",
"convertFloatToString": "CFT",
"convertStringToInt": "CTI",
"convertStringToSignedInt": "CTIS",
"convertStringToFloat": "CTF",
"writeString": "WRTT",
"equalString": "EQUT",
"findString": "FNDT",
"getStringLength": "LENT",
"getSubstring": "SUBT",
"concatenateString": "CONT",
"getSpaceDelimitedStringLength": "SDL",
"getSpaceDelimitedStringElement": "SDS",
"random": "RAN",
"sine": "SIN",
"writeTime": "TWR",
"readTime": "TRD",
"sleep": "TSL",
"clearDisplay": "DCL",
"writeDisplay": "DWR",
"readDisplay": "DRD",
"convertKeyToString": "CKT",
"convertStringToKey": "CTK",
"keyIsPressed": "KIP",
"promptKey": "PKY",
"promptString": "PTX",
"promptSelection": "PSL",
"promptFile": "PFL",
"setPortMode": "XSM",
"digitalReadPort": "XDR",
"digitalWritePort": "XDW",
"analogReadPort": "XAR",
"getNumberOfFiles": "RNF",
"openFileByIndex": "RGAI",
"openFileByName": "RGAN",
"openFileAfterFile": "RGAA",
"createFile": "RCR",
"readFile": "RRD",
"writeFile": "RWR",
"getFileName": "RGN",
"getFileSize": "RGS",
"deleteFile": "RDL",
"runFile": "RRN",
"quit": "QUIT",
"convertStringToOpcode": "CTO",
"convertByteToHex": "CIH",
"convertHexToByte": "CHI",
"allocate": "ALLO",
"free": "FREE",
"scrollDisplay": "DSC",
"getDisplayWidth": "DDX",
"getDisplayHeight": "DDY",
"gameOfLife": "GMLF",
"mandelbrot": "MNDL",
"pow": "POW",
"log": "LOG"
}

tempText = """void write(short len, data *src, data *dest) { }
void fill(int data, int2 amount, data *dest) { }
signed int2 find(int data, int2 len, data *src) { }
void convertIntToString(int num, byte *dest) { }
void convertSignedIntToString(signed int num, byte *dest) { }
void convertFloatToString(float num, byte *dest) { }
int convertStringToInt(byte *text) { }
signed int convertStringToSignedInt(byte *text) { }
float convertStringToFloat(byte *text) { }
void writeString(byte *text, byte *dest) { }
byte equalString(byte *text1, byte *text2) { }
signed int findString(byte *pattern, byte *text) { }
int getStringLength(byte *text) { }
void getSubstring(byte *text, int startIndex, int endIndex, byte *dest) { }
void concatenateString(byte *text, byte *dest) { }
int getSpaceDelimitedStringLength(byte *text) { }
void getSpaceDelimitedStringElement(byte *text, int index, byte *dest) { }
int random(int max) { }
float sine(float num) { }
void writeTime(int num) { }
int readTime() { }
void sleep(int num) { }
void clearDisplay() { }
void writeDisplay(int posX, int posY, byte *text) { }
void readDisplay(int posX, int posY, int2 len, byte *dest) { }
void convertKeyToString(short key, byte shift, byte *dest) { }
short convertStringToKey(byte *text) { }
byte keyIsPressed(short key) { }
short promptKey() { }
void promptString(int posY, byte *dest) { }
int promptSelection(int amount, byte *text) { }
int promptFile() { }
void setPortMode(byte mode, byte pin) { }
byte digitalReadPort(byte pin) { }
void digitalWritePort(byte value, byte pin) { }
short analogReadPort(byte pin) { }
int getNumberOfFiles() { }
signed long openFileByIndex(int index) { }
signed long openFileByName(byte *text) { }
signed long openFileAfterFile(signed long handle) { }
signed long createFile(byte *name, int size) { }
void readFile(signed long handle, byte *dest) { }
void writeFile(byte *src, signed long handle) { }
void getFileName(signed long handle, byte *dest) { }
int getFileSize(signed long handle) { }
void deleteFile(signed long handle) { }
void runFile(signed long handle) { }
void quit() { }
byte convertStringToOpcode(byte *text) { }
void convertByteToHex(byte num, byte *dest) { }
byte convertHexToByte(byte *text) { }
data *allocate(int len) { }
void free(data *ptr) { }
void scrollDisplay(byte direction, byte *text) { }
int getDisplayWidth() { }
int getDisplayHeight() { }
void gameOfLife(int width, int height, byte *src, byte *dest) { }
byte mandelbrot(float real, float imag, float offset, int iter) { }
float pow(float num, float exp) { }
float log(float num, float base) { }"""

tempLineList = tempText.split("\n")

index = 0
while index < len(tempLineList):
    tempLine = tempLineList[index]
    tempTokenList = convertCodeToTokenList(tempLine)
    tempTuple = getNextStatement(tempTokenList, 0)
    tempStatement = tempTuple[0]
    if not isinstance(tempStatement, FunctionDeclarationStatement):
        print "Bad bytecode function declaration:"
        print tempLine
        sys.exit()
    defaultFunctionDeclarationStatementSet.append(tempStatement)
    index += 1

def getDefaultFunctionDeclarationStatementByName(name):
    index = 0
    while index < len(defaultFunctionDeclarationStatementSet):
        tempFunctionDeclarationStatement = defaultFunctionDeclarationStatementSet[index]
        if tempFunctionDeclarationStatement.getName() == name:
            return tempFunctionDeclarationStatement
        index += 1
    return None

def populateLabelInvocations(bytecode, constantData):
    index = 0
    while index < len(labelStatementList):
        tempLabelStatement = labelStatementList[index]
        tempLabelStatement.populateLabelInvocations(bytecode, constantData)
        index += 1

precompilerDefinitionList = []

class PrecompilerDefinition:
    def __init__(self, name, value):
        self.name = name
        self.value = value
        precompilerDefinitionList.append(self)

def precompileCode(code):
    tempList = code.split("\n")
    output = ""
    index = 0
    while index < len(tempList):
        tempText = tempList[index]
        index2 = 0
        while index2 < len(tempText):
            tempCharacter = tempText[index2]
            if tempCharacter != " " and tempCharacter != "\t" and tempCharacter != "\n":
                break
            index2 += 1
        hasProcessedLine = False
        if index2 < len(tempText) - 8:
            if tempText[index2:(index2 + 8)] == "#define ":
                index3 = tempText.find(" ", index2 + 8)
                tempName = tempText[(index2 + 8):index3]
                tempValue = tempText[(index3 + 1):]
                PrecompilerDefinition(tempName, tempValue)
                hasProcessedLine = True
        if index2 < len(tempText) - 2:
            if tempText[index2:(index2 + 2)] == "//":
                hasProcessedLine = True
        if not hasProcessedLine:
            tempCount = 0
            hasPerformedSubstition = True
            while hasPerformedSubstition:
                hasPerformedSubstition = False
                if tempCount > 5000:
                    raise CodeException("Too many substitutions.", index)
                index2 = 0
                while index2 < len(precompilerDefinitionList):
                    tempPrecompilerDefinition = precompilerDefinitionList[index2]
                    tempName = tempPrecompilerDefinition.name
                    index3 = tempText.find(tempName)
                    if index3 > -1:
                        tempText = tempText[:index3] + tempPrecompilerDefinition.value + tempText[(index3 + len(tempName)):]
                        hasPerformedSubstition = True
                    index2 += 1
                tempCount += 1
            output += tempText
        output += "\n"
        index += 1
    return output

print "DUO Light Pseudo C Compiler"

if len(sys.argv) < 2:
    print "Please provide a file name as an argument."
    sys.exit(0)

inputFileName = sys.argv[1];
tempFile = open(inputFileName, "r")
inputText = tempFile.read()
tempFile.close()

tempBytecode = []
tempConstantData = []

try:
    tempText = precompileCode(inputText)
    tempTokenList = convertCodeToTokenList(tempText)
    tempStatement = convertTokenListToStatement(tempTokenList)
    tempStatement.resolveParentRelationships(None, None)
    tempStatement.addAuxiliaryLabelStatements(0)
    tempStatement.resolveDeclarations()
    tempStatement.determineVariableOffsets()
    tempBytecode.append(pointerSize)
    appendNumberToList(tempBytecode, pointerSize, 0)
    tempStatement.convertToBytecode(tempBytecode, tempConstantData)
    setNumberInList(tempBytecode, 1, pointerSize, tempStatement.getSize())
    tempStatement.addFunctionInfoToConstantData(tempConstantData)
    populateLabelInvocations(tempBytecode, tempConstantData)
except CodeException as exception:
    print exception.convertToText(inputText)
    sys.exit()
#print tempBytecode
#print tempConstantData
outputList = []
index = 0
while index < len(tempBytecode):
    tempNumber = tempBytecode[index]
    outputList.append(tempNumber)
    index += 1
index = 0
while index < len(tempConstantData):
    tempNumber = tempConstantData[index]
    outputList.append(tempNumber)
    index += 1

index = inputFileName.rfind(".")
outputFileName = inputFileName[:index] + ".dlbc"
tempFile = open(outputFileName, "wb")
tempFile.write(bytearray(outputList))
tempFile.close()

print "Finished compiling."
